//
//---------------------------------------------------------------------------
// Name    : HIDBootloader18F.bas
// Author  : Jerry Messina
// Date    : 2/1/2015
// Version : 1.2
// Notes   :
// This is a Swordfish port of the Microchip USB HID Bootloader for PIC18F
// It is a combination of the various seperate bootloaders into a single file
// that supports the entire 18F USB family, including the J and K series devices.
//
// Version 1.2
//  - add K50 and J94/J99
//  - incorporate usb v143 and misc fixes
//  - replace modified SF system.bas with new startup.bas module
// Version 1.1
//  - add code to clear additional registers before jumping to the application
//  - add system.bas v1.4 (thanks to RangerBob)
//
//---------------------------------------------------------------------------
//  Microchip USB HID Bootloader for PIC18F
// 
//  File version         Date        Comment
//  1.0                  06/19/08    Original Version.  Adapted from
//                                   MCHPFSUSB v2.1 HID Bootloader
//                                   for PIC18F87J50 Family devices.
//
// note: changes made in 2.9j (version "1.01") such as sign flash, extended query,
// and bootloader versioning have not been incorporated as they would make the 
// bootloader even larger
//---------------------------------------------------------------------------
//
program HIDBootloader18F

//
// Memory Requirements
//
// Flash Usage
// -----------
// SF USB HID bootloader requires almost the entire 4096 bytes of code space
// when compiled with all options and for the worst-case device, which is the
// K series parts. There is only 30 bytes or so of code space remaining, 
// so keep this in mind if you make any changes. If you need more space you 
// may have to disable some of the optional features, such as the IPC module
//
// RAM Usage
// ---------
// The program uses ~100 bytes for SF system vars + general variables
// Add to that 64 + 64 + 64 = 192 bytes for USB PacketFromPC, PacketToPC, and the
// ProgrammingBuffer gives total ram ~= 100 + 192 = 292 bytes (or thereabouts)
// Adding to that the USB HID ram requirements (approx 168 bytes) gives a total
// of about ~450 bytes min. 
//

//
//-----------------------------------------------------------------------------
// hardware platforms (uncomment one section/customize for your hardware)
// the USER_INIT file for each platform contains the pic config settings, 
// userInit() routine, LED definitions, and any misc setup info
//-----------------------------------------------------------------------------
//
{
// Sparkfun USB Bit Whacker v1.2 18F2553 DEV-00762 (www.sparkfun.com)
device = 18F2553        // 18F2550 family
clock = 48
#option USER_INIT = "UBW_2553.bas"
}

{
// Microchip LPC (LowPinCount) USB dev kit 18F14K50 DM164127
device = 18F14K50
clock = 48
#option USER_INIT = "LPC_14K50.bas"
}

{
// Microchip USB PIM 18F47J53 MA180029
device = 18F47J53
clock = 48
#option USER_INIT = "PIM_47J53.bas"
}

{
// Microchip USB PIM 18F87J94 MA180033
device = 18F87J94
clock = 64          // clock setup supports 64MHz or 32MHz
#option USER_INIT = "PIM_87J94.bas"
}

'{
// Microchip PICDEM FS USB dev board 18F45K50 DM163025-1
device = 18F45K50
clock = 48
#option USER_INIT = "PICDEM_FS_45K50.bas"
'}

//
//-----------------------------------------------------------------------------
// MPLAB ICD debugger support (true/false)
//-----------------------------------------------------------------------------
// note: if you enable the ICD and you get "Program code too large" errors
// you may need to increase the BOOTLOADER_SIZE and APP_PROGRAM_START settings.
// some older devices require a large amount of flash space for the debugger 
// executive (see the RESERVE_PAGES and RESERVE_BYTES definitions)
#option MPLAB_ICD = false

// enable SF diagnostics
'#option _showvar = true

//
//-----------------------------------------------------------------------------
// rom addresses and interrupt vector relocation
//-----------------------------------------------------------------------------
//
// bootloader max size
#option BOOTLOADER_SIZE = 4096

// application program start address
// **** THIS MUST MATCH THE APPLICATION '#option org_reset = xxxx' SETTING ****
// it should also match the _StartPageToErase_ definition later
#define APP_PROGRAM_START = $1000

// redirect interrupts to application code
#define VECTOR_HI_OFFSET = $0008
#define VECTOR_LO_OFFSET = $0018

#option vector_isr_hi = APP_PROGRAM_START + VECTOR_HI_OFFSET
#option vector_isr_lo = APP_PROGRAM_START + VECTOR_LO_OFFSET

// constrain ROM so that it doesn't exceed the bootloader size
// this will let the compiler tell us if the bootloader is too large
#variable _maxrom = BOOTLOADER_SIZE

// bring in the app program address as a const so we can jump to it
const APP_PROGRAM_ADDR as longword = APP_PROGRAM_START

//
//-----------------------------------------------------------------------------
// SF startup initialization
//-----------------------------------------------------------------------------
// note: include startup.bas first before any other files
// this is VERY important if you're using a bootloader
//
// set WDT option to allow clrwdt() to tickle the wdt (since it may be enabled), 
// but do not change the config setting
//
#option WDT = true
include "startup.bas"

//
//-----------------------------------------------------------------------------
// config settings
// Note: The bootloader provides the ability to reprogram the config settings,
// so it's possible to build the bootloader and the application with different
// settings. However, this can be very dangerous and can result in non-functional
// operation, or even a system unable to boot. It's probably best to just set 
// them once here, otherwise use great caution.
//-----------------------------------------------------------------------------
//
// include the USER_INIT file specified above
include USER_INIT

// debugger support file included after user init in case we need to override any 
// config settings the debugger may need (ie WDT off)
#if isoption(MPLAB_ICD) and (MPLAB_ICD)
  include "mplabicd.bas"
#endif

//
//-----------------------------------------------------------------------------
// IPC module (optional)
// used for application->bootloader communications
// If used, this file MUST be included before any other libraries/startup code
// that might modify the RCON register
//-----------------------------------------------------------------------------
//
#option USE_IPC = true
#if (USE_IPC)
  include "ipcBootloader.bas"
#endif

//
//-----------------------------------------------------------------------------
// User hardware routines
// **** customize to match your hardware ****
//-----------------------------------------------------------------------------
// the userio module maps some of the hardware defined in the USER_INIT
// file to some friendly names, including UserInit()/Exit(), UserBootselect(),
// and UserLedOn()/Off()/Toggle()
//
include "userio.bas"

//
//-----------------------------------------------------------------------------
// USB options and includes
//-----------------------------------------------------------------------------
//
// enable display of build status messages to the IDE Results window
#option SHOW_USB_BUILD_MSGS = false      // set false to disable messages, true to enable

// set USB options
#option HID_BUFFER_SIZE = 64

// these are set so that the Microchip HIDBootLoader.exe program can find the device
// **DO NOT CHANGE unless you recompile the HIDBootLoader to match**
#option USB_DESCRIPTOR = "HIDBootDescriptor.bas"    // name of descriptor file
#option USB_VID = $04D8                             // Vendor ID: Microchip (1240)
#option USB_PID = $003C                             // Product ID: HID Bootloader (60)

// do not let usbhid initialize itself... we'll call it if/when needed
#option USB_AUTO_INITIALIZE = false
// the bootloader CANNOT use interrupts **DO NOT CHANGE THIS**
#option USB_SERVICE = false
// save the code space... we'll use our own routine here
#option USB_CLOCK_USE_DELAYMS = false

// ensure we reserve 256 bytes of USB ram, and tell usbmem.bas
// to keep the entire block reserved and do not resize it
// see notes on ProgrammingBuffer() for the reasons why
#option USB_PROTECTED_BLOCKSIZE = 256
#option USBMEM_RESERVE_MEMORY = true

// import modules...
include "usbhid.bas"
include "usbsfr.bas"
include "usbsystem.bas"

// bring in protected region memory defines so we can inspect them
public const
    sf_protected_a_start = _protected_a_start,
    sf_protected_a_end   = _protected_a_end

//
//-----------------------------------------------------------------------------
// bootloader supported device memory definitions
//-----------------------------------------------------------------------------
//

// Beginning of application program memory (not occupied by bootloader)
#define _ProgramMemStart_ = APP_PROGRAM_START

// standard 18F series devices
#if _device in (18F2550, 18F4550)
    #define _EraseBlockSize_            = 64        //64 byte erase block
    #define _StartPageToErase_          = (_ProgramMemStart_/_EraseBlockSize_)  //application start page (64 x 64 = $1000)
    #define _MaxPageToErase_            = 511       //Last 64 byte page of flash
    #define _ProgramMemStop_            = $008000   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramBlockSize_          = $20       //32 byte programming block size
    #define _ConfigWordsStartAddress_   = $300000   //$300000 is CONFIG space
    #define _ConfigWordsSectionLength_  = 14        //14 bytes worth of Configuration words
    #define DEVICE_WITH_IDLOC
    #define _UserIDAddress_             = $200000   //User ID is 8 bytes starting at $200000
    #define _UserIDSize_                = 8
    #define DEVICE_WITH_EEPROM
    #define _EEPROMSize_                = $100      //256 bytes of EEPROM on this device
    #define _EEPROMEffectiveAddress_    = $F00000

#elseif _device in (18F2553, 18F4553)
    #define _EraseBlockSize_            = 64        //64 byte erase block
    #define _StartPageToErase_          = (_ProgramMemStart_/_EraseBlockSize_)  //application start page (64 x 64 = $1000)
    #define _MaxPageToErase_            = 511       //Last 64 byte page of flash
    #define _ProgramMemStop_            = $008000   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramBlockSize_          = $20       //32 byte programming block size
    #define _ConfigWordsStartAddress_   = $300000   //$300000 is CONFIG space
    #define _ConfigWordsSectionLength_  = 14        //14 bytes worth of Configuration words
    #define DEVICE_WITH_IDLOC
    #define _UserIDAddress_             = $200000   //User ID is 8 bytes starting at $200000
    #define _UserIDSize_                = 8
    #define DEVICE_WITH_EEPROM
    #define _EEPROMSize_                = $100      //256 bytes of EEPROM on this device
    #define _EEPROMEffectiveAddress_    = $F00000

#elseif _device in (18F2455, 18F4455)
    #define _EraseBlockSize_            = 64        //64 byte erase block
    #define _StartPageToErase_          = (_ProgramMemStart_/_EraseBlockSize_)  //application start page (64 x 64 = $1000)
    #define _MaxPageToErase_            = 383       //Last 64 byte page of flash
    #define _ProgramMemStop_            = $006000   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramBlockSize_          = $20       //32 byte programming block size
    #define _ConfigWordsStartAddress_   = $300000   //$300000 is CONFIG space
    #define _ConfigWordsSectionLength_  = 14        //14 bytes worth of Configuration words
    #define DEVICE_WITH_IDLOC
    #define _UserIDAddress_             = $200000   //User ID is 8 bytes starting at $200000
    #define _UserIDSize_                = 8
    #define DEVICE_WITH_EEPROM
    #define _EEPROMSize_                = $100      //256 bytes of EEPROM on this device
    #define _EEPROMEffectiveAddress_    = $F00000

#elseif _device in (18F2458, 18F4458)
    #define _EraseBlockSize_            = 64        //64 byte erase block
    #define _StartPageToErase_          = (_ProgramMemStart_/_EraseBlockSize_)  //application start page (64 x 64 = $1000)
    #define _MaxPageToErase_            = 383       //Last 64 byte page of flash
    #define _ProgramMemStop_            = $006000   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramBlockSize_          = $20       //32 byte programming block size
    #define _ConfigWordsStartAddress_   = $300000   //$300000 is CONFIG space
    #define _ConfigWordsSectionLength_  = 14        //14 bytes worth of Configuration words
    #define DEVICE_WITH_IDLOC
    #define _UserIDAddress_             = $200000   //User ID is 8 bytes starting at $200000
    #define _UserIDSize_                = 8
    #define DEVICE_WITH_EEPROM
    #define _EEPROMSize_                = $100      //256 bytes of EEPROM on this device
    #define _EEPROMEffectiveAddress_    = $F00000

#elseif _device in (18F2450, 18F4450)               // no EEPROM in 18Fx450
    #define _EraseBlockSize_            = 64        //64 byte erase block
    #define _StartPageToErase_          = (_ProgramMemStart_/_EraseBlockSize_)  //application start page (64 x 64 = $1000)
    #define _MaxPageToErase_            = 255       //Last 64 byte page of flash
    #define _ProgramMemStop_            = $004000   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramBlockSize_          = $10       //16 byte programming block size
    #define _ConfigWordsStartAddress_   = $300000   //$300000 is CONFIG space
    #define _ConfigWordsSectionLength_  = 14        //14 bytes worth of Configuration words
    #define DEVICE_WITH_IDLOC
    #define _UserIDAddress_             = $200000   //User ID is 8 bytes starting at $200000
    #define _UserIDSize_                = 8

//--------------------K series-------------------------------
#elseif _device in (18F14K50)
    #define DEVICE_K_SERIES
    #define _EraseBlockSize_            = 64        //64 byte erase block
    #define _StartPageToErase_          = (_ProgramMemStart_/_EraseBlockSize_)  //application start page (64 x 64 = $1000)
    #define _MaxPageToErase_            = 255       //Last 64 byte page of flash
    #define _ProgramMemStop_            = $004000   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramBlockSize_          = $10       //16 byte programming block size
    #define _ConfigWordsStartAddress_   = $300000   //$300000 is CONFIG space
    #define _ConfigWordsSectionLength_  = 14        //14 bytes worth of Configuration words
    #define DEVICE_WITH_IDLOC
    #define _UserIDAddress_             = $200000   //User ID is 8 bytes starting at $200000
    #define _UserIDSize_                = 8
    #define DEVICE_WITH_EEPROM
    #define _EEPROMSize_                = $100      //256 bytes of EEPROM on this device
    #define _EEPROMEffectiveAddress_    = $F00000

#elseif _device in (18F13K50)
    #define DEVICE_K_SERIES
    #define _EraseBlockSize_            = 64        //64 byte erase block
    #define _StartPageToErase_          = (_ProgramMemStart_/_EraseBlockSize_)  //application start page (64 x 64 = $1000)
    #define _MaxPageToErase_            = 127       //Last 64 byte page of flash
    #define _ProgramMemStop_            = $002000   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramBlockSize_          = $8        //8 byte programming block size
    #define _ConfigWordsStartAddress_   = $300000   //$300000 is CONFIG space
    #define _ConfigWordsSectionLength_  = 14        //14 bytes worth of Configuration words
    #define DEVICE_WITH_IDLOC
    #define _UserIDAddress_             = $200000   //User ID is 8 bytes starting at $200000
    #define _UserIDSize_                = 8
    #define DEVICE_WITH_EEPROM
    #define _EEPROMSize_                = $100      //256 bytes of EEPROM on this device
    #define _EEPROMEffectiveAddress_    = $F00000

#elseif _device in (18F24K50)
    #define DEVICE_K_SERIES
    #define _EraseBlockSize_            = 64        //64 byte erase block
    #define _StartPageToErase_          = (_ProgramMemStart_/_EraseBlockSize_)  //application start page (64 x 64 = $1000)
    #define _MaxPageToErase_            = 255       //Last 64 byte page of flash
    #define _ProgramMemStop_            = $004000   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramBlockSize_          = $40       //64 byte programming block size
    #define _ConfigWordsStartAddress_   = $300000   //$300000 is CONFIG space
    #define _ConfigWordsSectionLength_  = 14        //14 bytes worth of Configuration words
    #define DEVICE_WITH_IDLOC
    #define _UserIDAddress_             = $200000   //User ID is 8 bytes starting at $200000
    #define _UserIDSize_                = 8
    #define DEVICE_WITH_EEPROM
    #define _EEPROMSize_                = $100      //256 bytes of EEPROM on this device
    #define _EEPROMEffectiveAddress_    = $F00000

#elseif _device in (18F25K50, 18F45K50)
    #define DEVICE_K_SERIES
    #define _EraseBlockSize_            = 64        //64 byte erase block
    #define _StartPageToErase_          = (_ProgramMemStart_/_EraseBlockSize_)  //application start page (64 x 64 = $1000)
    #define _MaxPageToErase_            = 511       //Last 64 byte page of flash
    #define _ProgramMemStop_            = $008000   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramBlockSize_          = $40       //64 byte programming block size
    #define _ConfigWordsStartAddress_   = $300000   //$300000 is CONFIG space
    #define _ConfigWordsSectionLength_  = 14        //14 bytes worth of Configuration words
    #define DEVICE_WITH_IDLOC
    #define _UserIDAddress_             = $200000   //User ID is 8 bytes starting at $200000
    #define _UserIDSize_                = 8
    #define DEVICE_WITH_EEPROM
    #define _EEPROMSize_                = $100      //256 bytes of EEPROM on this device
    #define _EEPROMEffectiveAddress_    = $F00000

//--------------------J series-------------------------------
// J series devices do not have EEPROM or UserID's
#elseif _device in (18F67J50, 18F87J50)
    #define DEVICE_J_SERIES
    #define _EraseBlockSize_            = 1024      // 1024 byte erase block
    #define _StartPageToErase_          = (_ProgramMemStart_/_EraseBlockSize_)  //application start page (4 x 1024 = $1000)
    #define _MaxPageToEraseNoConfigs_   = 126       //Last page of flash without flash configuration words
    #define _MaxPageToEraseWithConfigs_ = 127       //Page 127 contains the flash configurations words
    #define _ProgramMemStopNoConfigs_   = $01FC00   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramMemStopWithConfigs_ = $01FFF8   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramBlockSize_          = $40       //64 byte flash block
    #define _ConfigWordsStartAddress_   = $01FFF8   //$XXXF8 is CONFIG1L
    #define _ConfigWordsSectionLength_  = $08       //8 bytes worth of Configuration words

#elseif _device in (18F66J50, 18F86J50)
    #define DEVICE_J_SERIES
    #define _EraseBlockSize_            = 1024      // 1024 byte erase block
    #define _StartPageToErase_          = (_ProgramMemStart_/_EraseBlockSize_)  //application start page (4 x 1024 = $1000)
    #define _MaxPageToEraseNoConfigs_   = 62        //Last page of flash without flash configuration words
    #define _MaxPageToEraseWithConfigs_ = 63        //Page 63 contains the flash configurations words
    #define _ProgramMemStopNoConfigs_   = $00FC00   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramMemStopWithConfigs_ = $00FFF8   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramBlockSize_          = $40       //64 byte flash block
    #define _ConfigWordsStartAddress_   = $00FFF8   //$XXXF8 is CONFIG1L
    #define _ConfigWordsSectionLength_  = $08       //8 bytes worth of Configuration words

#elseif _device in (18F66J55, 18F86J55)
    #define DEVICE_J_SERIES
    #define _EraseBlockSize_            = 1024      // 1024 byte erase block
    #define _StartPageToErase_          = (_ProgramMemStart_/_EraseBlockSize_)  //application start page (4 x 1024 = $1000)
    #define _MaxPageToEraseNoConfigs_   = 94        //Last page of flash without flash configuration words
    #define _MaxPageToEraseWithConfigs_ = 95        //Page 95 contains the flash configurations words
    #define _ProgramMemStopNoConfigs_   = $017C00   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramMemStopWithConfigs_ = $017FF8   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramBlockSize_          = $40       //64 byte flash block
    #define _ConfigWordsStartAddress_   = $017FF8   //$XXXF8 is CONFIG1L
    #define _ConfigWordsSectionLength_  = $08       //8 bytes worth of Configuration words

#elseif _device in (18F65J50, 18F85J50)
    #define DEVICE_J_SERIES
    #define _EraseBlockSize_            = 1024      // 1024 byte erase block
    #define _StartPageToErase_          = (_ProgramMemStart_/_EraseBlockSize_)  //application start page (4 x 1024 = $1000)
    #define _MaxPageToEraseNoConfigs_   = 30        //Last page of flash without flash configuration words
    #define _MaxPageToEraseWithConfigs_ = 31        //Page 31 contains the flash configurations words
    #define _ProgramMemStopNoConfigs_   = $007C00   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramMemStopWithConfigs_ = $007FF8   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramBlockSize_          = $40       //64 byte flash block
    #define _ConfigWordsStartAddress_   = $007FF8   //$XXXF8 is CONFIG1L
    #define _ConfigWordsSectionLength_  = $08       //8 bytes worth of Configuration words

#elseif _device in (18F24J50, 18F44J50)
    #define DEVICE_J_SERIES
    #define _EraseBlockSize_            = 1024      // 1024 byte erase block
    #define _StartPageToErase_          = (_ProgramMemStart_/_EraseBlockSize_)  //application start page (4 x 1024 = $1000)
    #define _MaxPageToEraseNoConfigs_   = 14        //Last page of flash without flash configuration words
    #define _MaxPageToEraseWithConfigs_ = 15        //Page 15 contains the flash configurations words
    #define _ProgramMemStopNoConfigs_   = $003C00   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramMemStopWithConfigs_ = $003FF8   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramBlockSize_          = $40       //64 byte flash block
    #define _ConfigWordsStartAddress_   = $003FF8   //$XXXF8 is CONFIG1L
    #define _ConfigWordsSectionLength_  = $08       //8 bytes worth of Configuration words

#elseif _device in (18F25J50, 18F45J50)
    #define DEVICE_J_SERIES
    #define _EraseBlockSize_            = 1024      // 1024 byte erase block
    #define _StartPageToErase_          = (_ProgramMemStart_/_EraseBlockSize_)  //application start page (4 x 1024 = $1000)
    #define _MaxPageToEraseNoConfigs_   = 30        //Last page of flash without flash configuration words
    #define _MaxPageToEraseWithConfigs_ = 31        //Page 31 contains the flash configurations words
    #define _ProgramMemStopNoConfigs_   = $007C00   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramMemStopWithConfigs_ = $007FF8   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramBlockSize_          = $40       //64 byte flash block
    #define _ConfigWordsStartAddress_   = $007FF8   //$XXXF8 is CONFIG1L
    #define _ConfigWordsSectionLength_  = $08       //8 bytes worth of Configuration words

#elseif _device in (18F26J50, 18F46J50)
    #define DEVICE_J_SERIES
    #define _EraseBlockSize_            = 1024      // 1024 byte erase block
    #define _StartPageToErase_          = (_ProgramMemStart_/_EraseBlockSize_)  //application start page (4 x 1024 = $1000)
    #define _MaxPageToEraseNoConfigs_   = 62        //Last page of flash without flash configuration words
    #define _MaxPageToEraseWithConfigs_ = 63        //Page 63 contains the flash configurations words
    #define _ProgramMemStopNoConfigs_   = $00FC00   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramMemStopWithConfigs_ = $00FFF8   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramBlockSize_          = $40       //64 byte flash block
    #define _ConfigWordsStartAddress_   = $00FFF8   //$XXXF8 is CONFIG1L
    #define _ConfigWordsSectionLength_  = $08       //8 bytes worth of Configuration words

#elseif _device in (18F26J53, 18F46J53)
    #define DEVICE_J_SERIES
    #define _EraseBlockSize_            = 1024      // 1024 byte erase block
    #define _StartPageToErase_          = (_ProgramMemStart_/_EraseBlockSize_)  //application start page (4 x 1024 = $1000)
    #define _MaxPageToEraseNoConfigs_   = 62        //Last page of flash without flash configuration words
    #define _MaxPageToEraseWithConfigs_ = 63        //Page 95 contains the flash configurations words
    #define _ProgramMemStopNoConfigs_   = $00FC00   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramMemStopWithConfigs_ = $00FFF8   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramBlockSize_          = $40       //64 byte flash block
    #define _ConfigWordsStartAddress_   = $00FFF8   //$XXXF8 is CONFIG1L
    #define _ConfigWordsSectionLength_  = $08       //8 bytes worth of Configuration words

#elseif _device in (18F27J53, 18F47J53)
    #define DEVICE_J_SERIES
    #define _EraseBlockSize_            = 1024      // 1024 byte erase block
    #define _StartPageToErase_          = (_ProgramMemStart_/_EraseBlockSize_)  //application start page (4 x 1024 = $1000)
    #define _MaxPageToEraseNoConfigs_   = 126       //Last page of flash without flash configuration words
    #define _MaxPageToEraseWithConfigs_ = 127       //Page 127 contains the flash configurations words
    #define _ProgramMemStopNoConfigs_   = $01FC00   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramMemStopWithConfigs_ = $01FFF8   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramBlockSize_          = $40       //64 byte flash block
    #define _ConfigWordsStartAddress_   = $01FFF8   //$XXXF8 is CONFIG1L
    #define _ConfigWordsSectionLength_  = $08       //8 bytes worth of Configuration words

#elseif _device in (18F97J94, 18F87J94, 18F67J94)   //128kB flash devices
    #define DEVICE_J_SERIES
    #define _EraseBlockSize_            = 512       // 512 byte erase block
    #define _StartPageToErase_          = (_ProgramMemStart_/_EraseBlockSize_)  //application start page (8 x 512 = $1000)
    #define _MaxPageToEraseNoConfigs_   = 254       //Last page of flash without flash configuration words
    #define _MaxPageToEraseWithConfigs_ = 255       //Page 255 contains the flash configurations words
    #define _ProgramMemStopNoConfigs_   = $01FE00   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramMemStopWithConfigs_ = $01FFF0   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramBlockSize_          = $40       //64 byte flash block
    #define _ConfigWordsStartAddress_   = $01FFF0   //$XXXF0 is CONFIG1L
    #define _ConfigWordsSectionLength_  = $10       //16 bytes worth of Configuration words

#elseif _device in (18F96J99, 18F86J99, 18F66J99)   //96kB flash devices
    #define DEVICE_J_SERIES
    #define _EraseBlockSize_            = 512       // 512 byte erase block
    #define _StartPageToErase_          = (_ProgramMemStart_/_EraseBlockSize_)  //application start page (8 x 512 = $1000)
    #define _MaxPageToEraseNoConfigs_   = 94        //Last page of flash without flash configuration words
    #define _MaxPageToEraseWithConfigs_ = 95        //Page 95 contains the flash configurations words
    #define _ProgramMemStopNoConfigs_   = $017C00   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramMemStopWithConfigs_ = $017FF0   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramBlockSize_          = $40       //64 byte flash block
    #define _ConfigWordsStartAddress_   = $01FFF0   //$XXXF0 is CONFIG1L
    #define _ConfigWordsSectionLength_  = $10       //16 bytes worth of Configuration words

#elseif _device in (18F96J94, 18F86J94, 18F66J94)   //64kB flash devices
    #define DEVICE_J_SERIES
    #define _EraseBlockSize_            = 512       // 512 byte erase block
    #define _StartPageToErase_          = (_ProgramMemStart_/_EraseBlockSize_)  //application start page (8 x 512 = $1000)
    #define _MaxPageToEraseNoConfigs_   = 62        //Last page of flash without flash configuration words
    #define _MaxPageToEraseWithConfigs_ = 63        //Page 63 contains the flash configurations words
    #define _ProgramMemStopNoConfigs_   = $00FC00   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramMemStopWithConfigs_ = $00FFF0   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramBlockSize_          = $40       //64 byte flash block
    #define _ConfigWordsStartAddress_   = $00FFF0   //$XXXF0 is CONFIG1L
    #define _ConfigWordsSectionLength_  = $10       //16 bytes worth of Configuration words

#elseif _device in (18F95J94, 18F85J94, 18F65J94)   //32kB flash devices
    #define DEVICE_J_SERIES
    #define _EraseBlockSize_            = 512       // 512 byte erase block
    #define _StartPageToErase_          = (_ProgramMemStart_/_EraseBlockSize_)  //application start page (8 x 512 = $1000)
    #define _MaxPageToEraseNoConfigs_   = 30        //Last page of flash without flash configuration words
    #define _MaxPageToEraseWithConfigs_ = 31        //Page 31 contains the flash configurations words
    #define _ProgramMemStopNoConfigs_   = $007C00   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramMemStopWithConfigs_ = $007FF0   //**MUST BE WORD ALIGNED (EVEN) ADDRESS
    #define _ProgramBlockSize_          = $40       //64 byte flash block
    #define _ConfigWordsStartAddress_   = $007FF0   //$XXXF0 is CONFIG1L
    #define _ConfigWordsSectionLength_  = $10       //16 bytes worth of Configuration words

#else
  #error "unsupported device"
#endif

// bring in #defines as constants so we can use them in executable statements
const
    _ProgramMemStart = _ProgramMemStart_

// if running under a debugger, we may have to adjust the upper program memory limits
// so we don't wipe out the debug executive when we erase/program
// this is not needed for the J series or the current K series devices (they use no rom)
#if (MPLAB_ICD and not defined(DEVICE_K_SERIES) and not defined(DEVICE_J_SERIES))
  const
      RESERVE_PAGES = 10,                 // just reserve ten 64 byte pages... that's the max used
      RESERVE_BYTES = RESERVE_PAGES * 64
#else
  const
      RESERVE_PAGES = 0,
      RESERVE_BYTES = 0
#endif

#if defined(DEVICE_J_SERIES)
const
    _EraseBlockSize             = _EraseBlockSize_,
    _StartPageToErase           = _StartPageToErase_,
    _StartEraseAddr             = _StartPageToErase_ * _EraseBlockSize_,
    _MaxPageToEraseNoConfigs    = _MaxPageToEraseNoConfigs_,
    _MaxPageToEraseWithConfigs  = _MaxPageToEraseWithConfigs_,
    _ProgramMemStopNoConfigs    = _ProgramMemStopNoConfigs_,
    _ProgramMemStopWithConfigs  = _ProgramMemStopWithConfigs_,
    _ProgramBlockSize           = _ProgramBlockSize_,
    _ConfigWordsStartAddress    = _ConfigWordsStartAddress_,
    _ConfigWordsSectionLength   = _ConfigWordsSectionLength_
#else   // standard 18F or K series
const
    _EraseBlockSize             = _EraseBlockSize_,
    _StartPageToErase           = _StartPageToErase_,
    _StartEraseAddr             = _StartPageToErase_ * _EraseBlockSize_,
    _MaxPageToErase             = _MaxPageToErase_ - RESERVE_PAGES,
    _ProgramMemStop             = _ProgramMemStop_ - RESERVE_BYTES,
    _ProgramBlockSize           = _ProgramBlockSize_,
    _ConfigWordsStartAddress    = _ConfigWordsStartAddress_,
    _ConfigWordsSectionLength   = _ConfigWordsSectionLength_
#endif  // DEVICE_J_SERIES

#if defined(DEVICE_WITH_IDLOC)
const
    _UserIDAddress  = _UserIDAddress_,
    _UserIDSize     = _UserIDSize_
#endif      // DEVICE_WITH_IDLOC

#if defined(DEVICE_WITH_EEPROM)
const
    _EEPROMSize             = _EEPROMSize_,
    _EEPROMEffectiveAddress = _EEPROMEffectiveAddress_
  // check EEPROM_SIZE
  #if (_EEPROMSize_ > 256)
    #error "code does not currently support devices with > 256 bytes of eeprom. must add EEADRH"
  #endif
#endif      // DEVICE_WITH_EEPROM

// sanity check some of the above settings
// application must start after (or equal) the BOOTLOADER_SIZE (typically they're the same)
#if (APP_PROGRAM_START < BOOTLOADER_SIZE)
  #error "**APP_PROGRAM_START must be located outside the bootloader memory space**"
// check starting erase block
#elseif (_ProgramMemStart_ < (_StartPageToErase_ * _EraseBlockSize_))
  #error "**app program will not be fully erased. check APP_PROGRAM_START and erase block definitions**"
// check app starting address boundary
#elseif ((_ProgramMemStart_ mod _EraseBlockSize_) <> 0)
  #error "**app program must begin on an _EraseBlockSize_ block boundary"
// check bootloader reserved space and erase blocks
#elseif (_StartPageToErase_ <> (_ProgramMemStart_/_EraseBlockSize_))
  #warning "**check erase block locations and definitions**"
#endif

//
//-----------------------------------------------------------------------------
// program variables, types , and constants
//-----------------------------------------------------------------------------
//

//Bootloader Commands
public const
    QUERY_DEVICE        = $02,  //Command that the host uses to learn about the device (what regions can be programmed, and what type of memory is the region)
    UNLOCK_CONFIG       = $03,  //Note, this command is used for both locking and unlocking the config bits (see the "//Unlock Configs Command Definitions" below)
    ERASE_DEVICE        = $04,  //Host sends this command to start an erase operation.  Firmware controls which pages should be erased.
    PROGRAM_DEVICE      = $05,  //If host is going to send a full REQUEST_DATABLOCK_SIZE to be programmed, it uses this command.
    PROGRAM_COMPLETE    = $06,  //If host send less than a REQUEST_DATABLOCK_SIZE to be programmed, or if it wished to program whatever was left in the buffer, it uses this command.
    GET_DATA            = $07,  //The host sends this command in order to read out memory from the device.  Used during verify (and read/export hex operations)
    RESET_DEVICE        = $08,  //Resets the microcontroller, so it can update the config bits (if they were programmed, and so as to leave the bootloader (and potentially go back into the main application)
    SIGN_FLASH          = $09,  //The host PC application should send this command after the verify operation has completed successfully.  If checksums are used instead of a true verify (due to ALLOW_GET_DATA_COMMAND being commented), then the host PC application should send SIGN_FLASH command after is has verified the checksums are as exected. The firmware will then program the SIGNATURE_WORD into flash at the SIGNATURE_ADDRESS.
    QUERY_EXTENDED_INFO = $0C   //Used by host PC app to get additional info about the device, beyond the basic NVM layout provided by the query device command

//Unlock Configs Command Definitions
public const
    UNLOCKCONFIG        = $00,  //Sub-command for the UNLOCK_CONFIG command
    LOCKCONFIG          = $01   //Sub-command for the UNLOCK_CONFIG command

//Query Device Response "Types"
//When the host sends a QUERY_DEVICE command, need to respond by populating a list of valid memory regions that exist in the device (and should be programmed)
public const
    MEMORY_REGION_PROGRAM_MEM   = $01,
    MEMORY_REGION_EEDATA        = $02,
    MEMORY_REGION_CONFIG        = $03,
    MEMORY_REGION_USERID        = $04,
    MEMORY_REGION_END           = $FF       // serves as a "list terminator" which denotes the end of the memory region list has been reached

const
    BOOTLOADER_V1_01_OR_NEWER_FLAG = $A5    //Tacked on in the VersionFlag byte, to indicate when using newer version of bootloader with extended query info available

//BootState Variable States
public const
    IDLE                = $00,
    NOT_IDLE            = $01,
    RUN_APPLICATION     = $02

//OtherConstants
const
    INVALID_ADDR        = $FFFFFFFF

//Application and Microcontroller constants
public const
    BYTES_PER_ADDRESS_PIC18 = $01,      //One byte per address.  PIC24 uses 2 bytes for each address in the hex file.
    WORDSIZE                = $02,      //PIC18 uses 2 byte words, PIC24 uses 3 byte words.
    USB_PACKET_SIZE         = 64,       // 64 byte packets
    PROG_BUF_SIZE           = 64,       // sizeof programming data buffer
    REQUEST_DATABLOCK_SIZE  = $3A       // Number of data bytes in a standard request to the PC.  Must be an even number from 2-58 ($02-$3A)
                                        // Larger numbers make better use of USB bandwidth and yield shorter program/verify times,
                                        // but require more RAM for buffer space

//General command packet structure used by PROGRAM_DEVICE and GET_DATA commands
structure TGENERAL_CMD
    Command     as byte
    Address     as longword
    Size        as byte
    Data(REQUEST_DATABLOCK_SIZE) as byte
end structure

//This struct used for responding to QUERY_DEVICE command (on a device with multiple programmable sections)
structure TQUERY_DEVICE
    Command             as byte
    PacketDataFieldSize as byte
    BytesPerAddress     as byte
    Type1               as byte             // TypeProgramMemory
    Address1            as longword
    Length1             as longword
    Type2               as byte             // TypeConfigWords
    Address2            as longword
    Length2             as longword
    Type3               as byte             // IDLOC (this is also 'EndOfTypes' for J series parts)
    Address3            as longword
    Length3             as longword
    Type4               as byte             // TypeEEPROM
    Address4            as longword
    Length4             as longword
    Type5               as byte             // TypeEndOfTypeList
    Address5            as longword
    Length5             as longword
    Type6               as byte
    Address6            as longword
    Length6             as longword
    VersionFlag         as byte             //Used by host software to identify if device is new enough to support QUERY_EXTENDED_INFO command
    ExtraPadBytes(6)    as byte
end structure

//For UNLOCK_CONFIG command
// ** added functionality **
// The standard Microchip HIDBootLoader.exe only allows for locking/unlocking the Config area.
// Here we add support for locking/unlocking the EEPROM and UserID areas as well so that you
// can do program updates without wiping out these areas.
// By default, the bootloader locks all of them, so if you use the standard HIDBootLoader.exe,
// programming will fail as the EEPROM and UserID areas will not get updated and they will fail
// verification. If you toggle the HIDBootLoader.exe 'Allow Configuration Word Programming'
// checkbox on and off, this will send unlock and lock packets which will clear the EEPROM lock
// and UserID lock bytes (since they're unused 00 bytes) and allow them to be updated on the next
// programming operation.
// If you wish to have individual control of each of the lock flags, you will have to modify the
// Microchip HIDBootLoader to add support for that (or use the updated v2.7.1 app)
//
structure TUNLOCK_CONFIG
    Command as byte
    ConfigLockValue as byte     // standard function
    EEPROMLockValue as byte     // added
    UserIDLockValue as byte     // added
end structure

structure TPC_PACKET
    contents(USB_PACKET_SIZE) as byte union
    b                         as TGENERAL_CMD union   // so we can access the .data() member array

    // common to all command packet types
    Command             as contents(0)

    // general command packet structure used by PROGRAM_DEVICE and GET_DATA commands
    Address             as contents(1).aslongword
    Size                as contents(5)
    // access the data() array using b.data() since we can't use an array here

    // QUERY_DEVICE
    PacketDataFieldSize as contents(1)
    BytesPerAddress     as contents(2)
    Type1               as contents(3)
    Address1            as contents(4).aslongword
    Length1             as contents(8).aslongword
    Type2               as contents(12)
    Address2            as contents(13).aslongword
    Length2             as contents(17).aslongword
    Type3               as contents(21)
    Address3            as contents(22).aslongword
    Length3             as contents(26).aslongword
    Type4               as contents(30)
    Address4            as contents(31).aslongword
    Length4             as contents(35).aslongword
    Type5               as contents(39)
    Address5            as contents(40).aslongword
    Length5             as contents(44).aslongword
    Type6               as contents(48)
    Address6            as contents(49).aslongword
    Length6             as contents(53).aslongword
    VersionFlag         as contents(57)

    // UNLOCK_CONFIG
    ConfigLockValue     as contents(1)     // standard
    EEPROMLockValue     as contents(2)     // added
    UserIDLockValue     as contents(3)     // added
end structure

structure T_LOCK_BITS
    b as byte
    ConfigLock  as b.booleans(0)
    EEPROMLock  as b.booleans(1)
    UserIDLock  as b.booleans(2)
end structure

//
// program variable declarations
//
public dim
    MaxPageToErase          as word,
    ProgramMemStopAddress   as longword,
    BootState               as byte,
    ErasePageTracker        as word,
    BufferedDataIndex       as byte,
    ProgrammedPointer       as longword,
    Locks                   as T_LOCK_BITS

public dim
    PacketFromPC    as TPC_PACKET,
    PacketToPC      as TPC_PACKET

// programming data buffer
// for most devices we could just let the buffer get allocated normally
// however, the 13K50 has such limited ram (256 bytes total) that we 
// have to steal part of the unused USB_RAM buffer from $2A8-$2E8
// this had to be done by hand... even freeing up enough space in the 
// _protected_a region wouldn't allow SF to allocate it automatically
// we could qualify this so as to only use the absolute method for the
// 13K50, but we might as well be consistant and do it the same for all
// devices... less to debug that way
public const PROG_BUF_ADDR = _protected_a_start + _HID_USED_RAM
public dim ProgrammingBuffer(PROG_BUF_SIZE) as byte absolute PROG_BUF_ADDR

// register helper definitions
public dim
    INTCON_GIE  as INTCON.7

// EECON1 bits
public dim
    EECON1_EEPGD as EECON1.7,      // flash program(=1) or data eeprom(=0) select
    EECON1_CFGS  as EECON1.6,      // configuration=1 (overrides EEPGD)
    EECON1_b5    as EECON1.5,      // unused
    EECON1_FREE  as EECON1.4,      // flash row erase enable=1, write=0
    EECON1_WRERR as EECON1.3,      // error flag
    EECON1_WREN  as EECON1.2,      // write enable=1
    EECON1_WR    as EECON1.1,      // initiate write=1
    EECON1_RD    as EECON1.0       // initiate eeprom read=1

//
//-----------------------------------------------------------------
// asm subs
//-----------------------------------------------------------------
//
// nop() - NOP instruction
public inline sub nop()
   asm
      nop
   end asm
end sub

// trap() - software breakpoint instruction
// this is undocumented... it assembles to an opcode of 0x00E0
// it works with pickit2 and other debuggers
public inline sub _trap()
  #if (MPLAB_ICD)
    asm
        trap
    end asm
  #endif
    // add a nop for brkpt skidding
    nop()
end sub

// reset() - 18F software Reset instruction
public inline sub reset()
  #if (MPLAB_ICD)
    // the RESET instruction doesn't work well when running under mplab...
    #warning "building for MPLAB ICD...replacing RESET with GOTO 0"
    INTCON = 0      // disable all intr
    _trap()         // breakpt so debugger knows we're resetting
    asm
        clrWDT
        goto 0
    end asm
  #else
    asm
        reset
    end asm
  #endif
end sub

// clrWDT() - clrwdt instruction
public inline sub clrWDT()
   asm
      clrWDT
   end asm
end sub

// table access instructions
inline sub TBLWT_()
    asm
        TBLWT*
    end asm
end sub

inline sub TBLWT_POSTINC()
    asm
        TBLWT*+
    end asm
end sub

inline function TBLRD_() as TABLAT
    asm
        TBLRD*
    end asm
end function

inline function TBLRD_POSTINC() as TABLAT
    asm
        TBLRD*+
    end asm
end function

inline function TBLRD_POSTDEC() as TABLAT
    asm
        TBLRD*-
    end asm
end function

//-----------------------------------------------------------------
// helper routines
//-----------------------------------------------------------------

//
// this routine provides a simple delay of about 50ms at 48MHz
// the delay period isn't critical, and will be longer at slower
// clock frequencies... it's just to allow the host a chance to
// detect us detaching before we reset. we could get a more accurate
// time using the builtin delayxx() routines, but that adds more code
//
sub LongDelay()
    dim i as word

    i = $ffff
    repeat
        clrWDT()
        nop()
        dec(i)
    until (i = 0)
end sub

//
// send unlock sequence and write data
//
sub UnlockAndActivate()
    // set unlock sequence and initiate WR
    EECON2 = $55
    EECON2 = $AA
    EECON1_WR = 1

    // wait until complete
    nop()
    while (EECON1_WR = 1)
    end while

    //clear the WREN bit for protection against self write/erase operations
    EECON1_WREN = 0
end sub

//
// in C18, TBLPTR is a 24-bit type that sets the TBLPTRU, TBLPTRH, and TBLPTRL
// here we create a sub to replicate that (a macro wouldn't work well here if
// an expression is used for the parameter)
//
sub TBLPTR(a as longword)
    TBLPTRU = a.bytes(2)
    TBLPTRH = a.bytes(1)
    TBLPTRL = a.bytes(0)
end sub

//
// The 18F TBLPTR is a 24-bit pointer. SF assumes that the upper byte (TBLPTRU)
// is always 0 and that the TBLPTR can be used as a 16-bit word. If you ever change
// TBLPTRU, you must set it back to 0 when you're done so that SF can access flash
// correctly (it uses the TBLPTR to access const data, which it locates in the first
// 64K of flash space).
// This should be called after using TBLPTR() to set an address
//
inline sub RestoreTBLPTR()
    TBLPTRU = 0
end sub

// read a byte of flash memory at 'addr' (assumes TBLPTRU is set correctly)
inline function ReadROM(addr as TABLEPTR) as TABLAT
    TBLRD_POSTINC()
end function

//
// erases a page of flash memory
// set ErasePageTracker prior to call
//
inline sub EraseFlashPage()
    // convert page number to linear address
  #if (_EraseBlockSize_ = 64)
    TBLPTR(longword(ErasePageTracker) << 6)     // a word would be ok, but this allows for mem > 64K
  #elseif (_EraseBlockSize_ = 512)
    TBLPTR(longword(ErasePageTracker) << 9)
  #elseif (_EraseBlockSize_ = 1024)
    TBLPTR(longword(ErasePageTracker) << 10)
  #else
    #error "unsupported _EraseBlockSize"
  #endif

    // prepare for erasing flash memory
  #if defined(DEVICE_J_SERIES)
    EECON1 = %00010100
  #else
    EECON1 = %10010100
  #endif

    UnlockAndActivate()
    RestoreTBLPTR()
end sub

//
// write blocks of data to flash
//
#if defined(DEVICE_J_SERIES)
// J series write (use word writes to write code chunks less than a full 64 byte block size)
sub WriteFlashSubBlock()
    FSR0 = addressof(ProgrammingBuffer)
    while (BufferedDataIndex > 0)        //While data is still in the buffer
        TBLPTR(ProgrammedPointer - BufferedDataIndex)
        //Below section will need to be modified if the WORDSIZE of your processor is not 2 bytes
        TABLAT = POSTINC0
        TBLWT_POSTINC()
        TABLAT = POSTINC0
        TBLWT_()                 //Do not increment TBLPTR on the second write.  See datasheet

        EECON1 = %00100100      //Word programming mode
        UnlockAndActivate()

        BufferedDataIndex = BufferedDataIndex - WORDSIZE   //Used up one word from the buffer
        clrWDT()
    end while
    RestoreTBLPTR()
end sub
#else       // not defined(DEVICE_J_SERIES)
// regular write routine (standard 18F and K series)
sub WriteFlashBlock()
    dim
        i as byte,
        BytesTakenFromBuffer as byte,
        CorrectionFactor as byte

    BytesTakenFromBuffer = 0
    TBLPTR(ProgrammedPointer - BufferedDataIndex)

    // comment from the original C source:
    // Check the lower 5 bits of the TBLPTR to verify it is pointing to a 32 byte aligned block (5 LSb = 00000).
    // If it isn't, need to somehow make it so before doing the actual loading of the programming latches.
    // In order to maximize programming speed, the PC application meant to be used with this firmware will not
    // send large blocks of 0xFF bytes.  If the PC application detects a large block of unprogrammed space in
    // the hex file (effectively = 0xFF), it will skip over that section and will not send it to the firmware.
    // This works, because the firmware will have already done an erase on that section of memory when it received
    // the ERASE_DEVICE command from the PC.  Therefore, the section can be left unprogrammed (after an erase
    // the flash ends up = 0xFF). This can result in a problem however, in that the next genuine non-0xFF section
    // in the hex file may not start on a 32 byte aligned block boundary.  This needs to be handled with care
    // since the microcontroller can only program 32 byte blocks that are aligned with 32 byte boundaries.
    // So, use the below code to avoid this potential issue.

    #if (_ProgramBlockSize_ = $40)
        CorrectionFactor = TBLPTRL and %00111111  //Correctionfactor = number of bytes tblptr must go back to find the immediate preceeding 32 byte boundary
        TBLPTRL = TBLPTRL and %11000000           //Move the table pointer back to the immediately preceeding 64 byte boundary
    #elseif (_ProgramBlockSize_ = $20)
        CorrectionFactor = TBLPTRL and %00011111  //Correctionfactor = number of bytes tblptr must go back to find the immediate preceeding 32 byte boundary
        TBLPTRL = TBLPTRL and %11100000           //Move the table pointer back to the immediately preceeding 32 byte boundary
    #elseif (_ProgramBlockSize_ = $10)
        CorrectionFactor = TBLPTRL and %00001111  //Correctionfactor = number of bytes tblptr must go back to find the immediate preceeding 16 byte boundary
        TBLPTRL = TBLPTRL and %11110000           //Move the table pointer back to the immediately preceeding 16 byte boundary
    #elseif (_ProgramBlockSize_ = $08)
        CorrectionFactor = TBLPTRL and %00000111  //Correctionfactor = number of bytes tblptr must go back to find the immediate preceeding 8 byte boundary
        TBLPTRL = TBLPTRL and %11111000           //Move the table pointer back to the immediately preceeding 8 byte boundary
    #else
        #error "unsupported _ProgramBlockSize"
    #endif

    i = 0
    while (i < _ProgramBlockSize)               //Load the programming latches
        if (CorrectionFactor = 0) then
            if (BufferedDataIndex <> 0) then    //If the buffer isn't empty
                TABLAT = ProgrammingBuffer(BytesTakenFromBuffer)
                TBLWT_POSTINC()
                inc(BytesTakenFromBuffer)
                dec(BufferedDataIndex)          //Used up a byte from the buffer
            else    //No more data in buffer, need to write 0xFF to fill the rest of the page
                TABLAT = $FF
                TBLWT_POSTINC()
            endif
        else
            TABLAT = $FF
            TBLWT_POSTINC()
            dec(CorrectionFactor)
        endif
        clrWDT()
        i = i + 1
    end while

    //Need to make table pointer point to the region which will be programmed before initiating the programming operation
    TBLRD_POSTDEC()         //Do this instead of TBLPTR-- since it takes less code space

    EECON1 = %10100100      // flash programming mode
    UnlockAndActivate()

    //Now need to fix the ProgrammingBuffer[].  We may not have taken a full 32 bytes out of the buffer.
    //In this case, the data is no longer justified correctly, and we need to rejustify the remaining
    //data to the "left" of the buffer (if there is any left)
    FSR0 = addressof(ProgrammingBuffer)     // for ProgrammingBuffer(i)
    FSR1 = addressof(ProgrammingBuffer)     // for ProgrammingBuffer(BytesTakenFromBuffer+i)
    inc(FSR1, BytesTakenFromBuffer)
    i = 0
    while (i < BufferedDataIndex)
        POSTINC0 = POSTINC1         //ProgrammingBuffer(i) = ProgrammingBuffer(BytesTakenFromBuffer+i)
        inc(i)
    end while
    RestoreTBLPTR()
end sub
#endif      // DEVICE_J_SERIES

//
// write Configuration data (also used to write the UserID locations)
//
sub WriteConfigBits()
    dim
        i as byte,
        ix as byte

  #if (MPLAB_ICD)
    '_trap()                 // for debugger breakpoint
    exit                    // just skip this if debugging (otherwise it'll likely crash)
  #endif

    TBLPTR(PacketFromPC.Address)

    ix = REQUEST_DATABLOCK_SIZE - PacketFromPC.Size
    FSR0 = addressof(PacketFromPC.b.Data)   // FSR = @PacketFromPC.b.Data(ix)
    inc(FSR0, ix)
    i = 0
    while (i < PacketFromPC.Size)
        TABLAT = POSTINC0                   // PacketFromPC.b.Data(ix)
        TBLWT_()
        EECON1 = %11000100                  // Config bits programming mode
        UnlockAndActivate()                 // write data
        TBLRD_POSTINC()                     // use this to incr tableptr after data's been written
        inc(i)
    end while
    RestoreTBLPTR()
end sub

//
// write eeprom data
//
#if defined(DEVICE_WITH_EEPROM)
sub WriteEEPROM()
    dim
        i as byte,
        ix as byte

    ix = REQUEST_DATABLOCK_SIZE - PacketFromPC.Size
    FSR0 = addressof(PacketFromPC.b.Data)   // FSR = @PacketFromPC.b.Data(ix)
    inc(FSR0, ix)
    EEADR = byte(PacketFromPC.Address)
    i = 0
    while (i < PacketFromPC.Size)
        EEDATA = POSTINC0       // PacketFromPC.b.Data(ix)
        EECON1 = %00000100      // EEPROM Write mode
        UnlockAndActivate()
        inc(EEADR)              // next EEADR
        inc(i)
    end while
end sub
#endif      // DEVICE_WITH_EEPROM

//
// erase UserID locations (0x200000 to 0x200007)
// note: this will erase the entire page, so it will also erase the
// debugger bkbgvectmem (region=0x200028-0x200037), which will crash
// the icd unless code is added to save/restore the vectors
//
#if defined(DEVICE_WITH_IDLOC)
sub EraseUserID()
  #if (MPLAB_ICD)
    '_trap()                 // for debugger breakpoint
    exit                    // just skip this if debugging
  #endif

    TBLPTR(_UserIDAddress)
    EECON1 = %10010100      // prepare for erasing flash memory
    UnlockAndActivate()
    RestoreTBLPTR()
end sub
#endif      //DEVICE_WITH_IDLOC

//
//-----------------------------------------------------------------------------
// void ProcessIO(void)
// This function processes the USB packets
//-----------------------------------------------------------------------------
//
public sub ProcessIO()
    dim
        i as byte,
        ix as byte

    if (BootState = IDLE) then
        if (mHIDRxIsBusy = 0) then     //did we receive a command?
            HIDRxReport(addressof(PacketFromPC), USB_PACKET_SIZE)
            BootState = NOT_IDLE
            UserLedOn()                 // flash user LED on and off as we execute packets
            // prepare the next packet we will send to the host by initializing the entire packet to 0x00
            clear(PacketToPC.Contents)
        else
            UserLedOff()
        endif
    else    // BootState must be NOT_IDLE...
        select (PacketFromPC.Command)
            case QUERY_DEVICE
                //Prepare a response packet, which lets the PC software know about the memory ranges of this device
                PacketToPC.Command = QUERY_DEVICE
                PacketToPC.PacketDataFieldSize = REQUEST_DATABLOCK_SIZE
                PacketToPC.BytesPerAddress = BYTES_PER_ADDRESS_PIC18
                PacketToPC.Type1 = MEMORY_REGION_PROGRAM_MEM
                PacketToPC.Address1 = _ProgramMemStart
                PacketToPC.Length1 = ProgramMemStopAddress - _ProgramMemStart  //Size of program memory area
                PacketToPC.Type2 = MEMORY_REGION_CONFIG
                PacketToPC.Address2 = _ConfigWordsStartAddress
                PacketToPC.Length2 = _ConfigWordsSectionLength

                // J series marks the end here (others may over-write end of list...)
                PacketToPC.Type3 = MEMORY_REGION_END
                PacketToPC.VersionFlag = 0      // BOOTLOADER_V1_01_OR_NEWER_FLAG not supported
                
              #if defined(DEVICE_WITH_IDLOC)
                PacketToPC.Type3 = MEMORY_REGION_PROGRAM_MEM       //Not really program memory (User ID), but the host treats it as such
                PacketToPC.Address3 = _UserIDAddress
                PacketToPC.Length3 = _UserIDSize
                PacketToPC.Type4 = MEMORY_REGION_END
              #endif

              #if defined(DEVICE_WITH_EEPROM)
                PacketToPC.Type4 = MEMORY_REGION_EEDATA
                PacketToPC.Address4 = _EEPROMEffectiveAddress
                PacketToPC.Length4 = _EEPROMSize
                PacketToPC.Type5 = MEMORY_REGION_END
              #endif

                // send the info to the host
                if (mHIDTxIsBusy = 0) then
                    HIDTxReport(addressof(PacketToPC), USB_PACKET_SIZE)
                    BootState = IDLE
                endif

            case UNLOCK_CONFIG
                // assume all three areas are locked (config, EEPROM, and UserID)...
                Locks.b = %111
              #if defined(DEVICE_J_SERIES)
                MaxPageToErase = _MaxPageToEraseNoConfigs
                ProgramMemStopAddress = _ProgramMemStopNoConfigs
                if (PacketFromPC.ConfigLockValue = UNLOCKCONFIG) then
                    Locks.ConfigLock = false
                    MaxPageToErase = _MaxPageToEraseWithConfigs
                    ProgramMemStopAddress = _ProgramMemStopWithConfigs
                endif
              #else
                MaxPageToErase = _MaxPageToErase
                ProgramMemStopAddress = _ProgramMemStop
                if (PacketFromPC.ConfigLockValue = UNLOCKCONFIG) then
                    Locks.ConfigLock = false
                endif
                if (PacketFromPC.EEPROMLockValue = UNLOCKCONFIG) then
                    Locks.EEPROMLock = false
                endif
                if (PacketFromPC.UserIDLockValue = UNLOCKCONFIG) then
                    Locks.UserIDLock = false
                endif
              #endif
                BootState = IDLE

            case ERASE_DEVICE
                // first erase main program flash memory
                for ErasePageTracker = _StartPageToErase to MaxPageToErase
                    EraseFlashPage()
                    clrWDT()
                    USBDriverService()     //Call USBDriverService() periodically
                next

              #if defined(DEVICE_WITH_EEPROM)           // device has EEPROM
                if (Locks.EEPROMLock = false) then      // is eeprom unlocked?
                    for EEADR = 0 to (_EEPROMSize-1)    // erase EEPROM
                        EEDATA = $FF
                        EECON1 = %00000100              //EEPROM Write mode
                        UnlockAndActivate()
                        clrWDT()
                        USBDriverService()              //Call USBDriverService() periodically
                    next
                endif
              #endif

              #if defined(DEVICE_WITH_IDLOC)            // device has UserID
                if (Locks.UserIDLock = false) then      // is UserID unlocked?
                    EraseUserID()                       // erase UserID
                endif
              #endif

                BootState = IDLE

            case PROGRAM_DEVICE
                // check if host is trying to program the config bits (0x30 implies config bits)
                if (PacketFromPC.Contents(3) = $30) then    //PacketFromPC.Contents[3] is MSB of the address
                    // config doesn't get reprogrammed unless the UNLOCK_CONFIG (LockValue = UNLOCKCONFIG) command has been sent
                    if (Locks.ConfigLock = false) then
                        WriteConfigBits()
                    endif
                    BootState = IDLE
                    exit
                endif

              #if defined(DEVICE_WITH_EEPROM)
                // check if host is trying to program the EEPROM (0xF0 implies EEPROM)
                if (PacketFromPC.Contents(3) = $F0) then    //PacketFromPC.Contents[3] is MSB of the address
                    if (Locks.EEPROMLock = false) then
                        WriteEEPROM()
                    endif
                    BootState = IDLE
                    exit
                endif
              #endif

              #if defined(DEVICE_WITH_IDLOC)
                // check if host is trying to program the UserID (0x20 implies UserID)
                if (PacketFromPC.Contents(3) = $20) then    //PacketFromPC.Contents[3] is MSB of the address
                    if (Locks.UserIDLock = true) then       // skip it... otherwise it'll get programmed below
                        BootState = IDLE
                        exit
                    endif
                endif
              #endif

                if (ProgrammedPointer = INVALID_ADDR) then
                    ProgrammedPointer = PacketFromPC.Address
                endif

                if (ProgrammedPointer = PacketFromPC.Address) then
                    i = 0
                    ix = REQUEST_DATABLOCK_SIZE - PacketFromPC.Size
                    while (i < PacketFromPC.Size)
                        // data field is right justified... need to put it in the buffer left justified
                        ProgrammingBuffer(BufferedDataIndex) = PacketFromPC.b.Data(ix)
                        inc(ix)
                        inc(BufferedDataIndex)
                        inc(ProgrammedPointer)

                      #if defined(DEVICE_J_SERIES)
                        if (BufferedDataIndex = REQUEST_DATABLOCK_SIZE) then
                            WriteFlashSubBlock()
                        endif
                      #else
                        if (BufferedDataIndex = _ProgramBlockSize) then
                            WriteFlashBlock()
                        endif
                      #endif
                        inc(i)
                    end while
                endif
                // else host sent us a non-contiguous packet address...  to make this firmware simpler,
                // host should not do this without sending a PROGRAM_COMPLETE command in between program sections
                BootState = IDLE

            case PROGRAM_COMPLETE
              #if defined(DEVICE_J_SERIES)
                WriteFlashSubBlock()
              #else
                WriteFlashBlock()
              #endif
                //Reinitialize pointer to an invalid range, so we know the next PROGRAM_DEVICE
                //will be the start address of a contiguous section
                ProgrammedPointer = INVALID_ADDR
                BootState = IDLE

            case GET_DATA
                PacketToPC.Command = GET_DATA
                PacketToPC.Address = PacketFromPC.Address
                PacketToPC.Size = PacketFromPC.Size

                TBLPTR(PacketFromPC.Address)
                i = 0
                ix = (USB_PACKET_SIZE - 6) - PacketFromPC.Size    // data index
                FSR0 = addressof(PacketToPC.b.Data)                 // FSR = PacketToPC.b.Data(ix)
                inc(FSR0, ix)
                while (i < PacketFromPC.Size)
                    if (PacketFromPC.Contents(3) = $F0) then        //0xF0 implies EEPROM, which doesn't use the table pointer
                      #if defined(DEVICE_WITH_EEPROM)
                        EEADR = byte(PacketFromPC.Address) + i      //The addr LSB is mapped 1:1 to the EEPROM address space
                        EECON1 = %00000000    //EEPROM read mode
                        EECON1_RD = 1
                        POSTINC0 = EEDATA
                      #else
                        nop()                   // it's good to have SOMETHING here...
                      #endif
                    else    //else must have been a normal program memory region, or one that can be read from with the table pointer
                        TBLRD_POSTINC()                     // TABLAT now has the value and TBLPTR has been incremented

                      #if not defined(DEVICE_J_SERIES)      // only for standard 18F and 18FxxK
                        // since 0x300004 (CONFIG3L) and 0x300007 (CONFIG4H) are not implemented,
                        // we need to return 0xFF since the device reads 0x00 but the hex file has 0x00 (??)
                        if (TBLPTRU = $30) then
                            if (TBLPTRL = $05) then     // check for 0x300004 (tblptr postinc)
                                TABLAT = $FF
                            endif
                            if (TBLPTRL = $08) then     // check for 0x300007 (tblptr postinc)
                                TABLAT = $FF
                            endif
                        endif
                      #endif
                        POSTINC0 = TABLAT
                    endif
                    inc(i)
                end while

                RestoreTBLPTR()

                if (mHIDTxIsBusy = 0) then
                    HIDTxReport(addressof(PacketToPC), USB_PACKET_SIZE)
                    BootState = IDLE
                endif

            case RESET_DEVICE
                USBSoftDetach()     // disable USB module
                // wait a while for the USB cable capacitance to discharge down to disconnected (SE0) state
                // otherwise host might not realize we disconnected/reconnected when we do the reset
                LongDelay()
                UserExit()
                Reset()
        end select
    endif   //BootState = IDLE

end sub

//
//-----------------------------------------------------------------------------
// application code jump (normal startup)
//-----------------------------------------------------------------------------
//
inline sub RunApplication()
    // we could check here to see if there was a valid application in memory
    // but since there isn't enough program space to do any kind of CRC check
    // or validation we'll just assume it's ok. An empty device will execute
    // as nops, so we'll get back here eventually if that's the case...
    // we will do one simple check, and that's to see if the 1st byte of the
    // app code is blank (0xff). if it is, then we'll just return
    RestoreTBLPTR()
    if (ReadROM(APP_PROGRAM_ADDR) = $ff) then
        exit
    endif

    // restore any user hardware settings
    UserExit()

    // set TBLPTR as it normally would be
    TBLPTRU = $00
    TBLPTRH = $00
    TBLPTRL = $00

    // set other important registers to their default state
    startup.SetRegisterDefaults()
        
    // reset wdt
    clrWDT()

    // here we go...
    asm
        goto APP_PROGRAM_ADDR
    end asm
end sub

//
//-----------------------------------------------------------------------------
// InitBootloader() - initialize bootloader state variables
//-----------------------------------------------------------------------------
//
inline sub InitBootloader()
    // assume we will not allow erase/programming of any of the special areas
    // (config, EEPROM, and UserID) unless host sends override command
    Locks.ConfigLock = true
    Locks.EEPROMLock = true
    Locks.UserIDLock = true

  #if defined(DEVICE_J_SERIES)
    MaxPageToErase = _MaxPageToEraseNoConfigs
    ProgramMemStopAddress = _ProgramMemStopNoConfigs
  #else
    MaxPageToErase = _MaxPageToErase
    ProgramMemStopAddress = _ProgramMemStop
  #endif

    BootState = IDLE
    ProgrammedPointer = INVALID_ADDR
    BufferedDataIndex = 0
end sub

//
//-----------------------------------------------------------------------------
// initialize USB and bootloader
//-----------------------------------------------------------------------------
//
sub InitializeSystem()
    // we should have gotten here from a RESET, but if not then the USB module may
    // have already been on if the application firmware calls the bootloader without
    // first disabling the USB module. If this happens, we need to soft-detach from
    // the host and wait a delay to allow host software to recognize the detach
    // before we turn USB back on
    if (UCONbits.USBEN = 1) then
        USBSoftDetach()     // disable USB module
        LongDelay()
    endif

    // init usb HID
    HID.Initialize()
    
    // we may need to wait for the usb hdw (clocks, etc) to stabilize
    LongDelay()

    // initialize bootloader program state
    InitBootloader()
end sub

//
//-------------------------------------------------------------
// main entry point
//-------------------------------------------------------------
//
main:
    // interrupts should already be disabled, but just in case...
    INTCON_GIE = 0

    // init user-defined hardware (leds, switch, etc)
    UserInit()

    // check Bootload Mode Entry Condition
    BootState = RUN_APPLICATION         // assume we're going to run the application
    // check switch/IO pin...
    if (UserBootselect() = true) then
        BootState = IDLE                // user wants to stay here
    endif
  #if (USE_IPC)
    // check to see if we were called from the application program...
    if (_rcon.RI = 0) then              // software RESET instruction detected...
        if (ipc_request = REQUEST_BOOTLOADER) then
            BootState = IDLE            // user wants to stay here
        endif
    endif
    IPC_REG = 0                         // clear register used for app->boot comm
  #endif

    // if there's no reason to stay, then run the application
    if (BootState = RUN_APPLICATION) then
        RunApplication()
    endif

    // setup bootloader
    InitializeSystem()

    // wait to connect to USB...
    UserLedOn()             // turn on led to indicate we're here waiting to enumerate...
    repeat
        USBservice()
    until (HID.Attached)
    UserLedOff()            // turn it off now that we're attached

    // we're connected. loop waiting for commands from the host
    while (true)
        clrWDT()
        USBservice()        // Need to call USBservice() periodically to handle USB
        ProcessIO()         // handle any data transfer/self programming
    end while

end program
