//
// main test program application for the usb hid bootloader
//
// at reset the bootloader passes control to this program, which sits in an
// infinite loop toggling an led every 500ms waiting for a switch to be pressed.
// once a switch press is detected, the led flashes rapidly waiting for the user
// to release the switch, at which point it generates a software reset, the pic
// restarts and the bootloader should run again starting the process all over.
//
program test_program

//
//-----------------------------------------------------------------------------
// hardware platforms (uncomment one section/customize for your hardware)
// the USER_INIT file for each platform contains the pic config settings, 
// userInit() routine, LED definitions, and any misc setup info
//-----------------------------------------------------------------------------
//
{
// Sparkfun USB Bit Whacker v1.2 18F2553 DEV-00762 (www.sparkfun.com)
device = 18F2553        // 18F2550 family
clock = 48
#option USER_INIT = "UBW_2553.bas"
}

{
// Microchip LPC (LowPinCount) USB dev kit 18F14K50 DM164127
device = 18F14K50
clock = 48
#option USER_INIT = "LPC_14K50.bas"
}

{
// Microchip USB PIM 18F47J53 MA180029
device = 18F47J53
clock = 48
#option USER_INIT = "PIM_47J53.bas"
// J series doesn't support these features
#option INCLUDE_USERID_DATA = false
#option INCLUDE_EEPROM_DATA = false
}

'{
// Microchip USB PIM 18F87J94 MA180033
device = 18F87J94
clock = 64          // clock setup supports 64MHz or 32MHz
#option USER_INIT = "PIM_87J94.bas"
// J series doesn't support these features
#option INCLUDE_USERID_DATA = false
#option INCLUDE_EEPROM_DATA = false
'}

{
// Microchip PICDEM FS USB dev board 18F45K50 DM163025-1
device = 18F45K50
clock = 48
#option USER_INIT = "PICDEM_FS_45K50.bas"
}

//
//-----------------------------------------------------------------------------
// program build options (ORG_RESET and/or ORG_PROGRAM) and ICD debugger
//-----------------------------------------------------------------------------
// BUILD_TYPE lets you change the environment the app program is built for
// note: normally, you can replace this entire section with:
//   #define APP_PROGRAM_START = $1000     // MUST MATCH THE BOOTLOADER SETTING
//   #option org_reset = APP_PROGRAM_START
// and that will set the application for use w/the HID bootloader
// 
// uncomment one build option
#option BUILD_TYPE = BOOTLOADER_BUILD      // no ICD, run w/bootloader (normal app w/bootloader)
'#option BUILD_TYPE = STANDALONE_BUILD      // no ICD, no bootloader (use MPLAB Programmer)
'#option BUILD_TYPE = DEBUG_BUILD           // run w/ICD, no bootloader (use MPLAB Debugger)

// process BUILD_TYPE option
#if (BUILD_TYPE = DEBUG_BUILD)
  #warning "building for MPLAB ICD operation (no bootloader)"
  #option MPLAB_ICD = true
  #option BUILD_STANDALONE = true
#elseif (BUILD_TYPE = STANDALONE_BUILD)
  #warning "building for STANDALONE operation (no bootloader)"
  #option MPLAB_ICD = false
  #option BUILD_STANDALONE = true
#elseif (BUILD_TYPE = BOOTLOADER_BUILD)
  // building for the HID bootloader
  #option MPLAB_ICD = false
  #option BUILD_STANDALONE = false
#else
  #error "invalid BUILD_TYPE"
#endif

// to build an application for use with a bootloader, the default is not to use either option
#option MPLAB_ICD = false
#option BUILD_STANDALONE = false

// application program start address
// ** THIS MUST MATCH THE BOOTLOADER SETTING **
#define APP_PROGRAM_START = $1000     // 4096

// locate the main app code at the same address no matter how we build
// this way we can see where things are going to end up
// (with the possible exception of the reset and intr vector jumps)
#if (BUILD_STANDALONE = true)
  // not using a bootloader
  // leave the reset and int vectors at their normal addr (0x0000, 0x0008, and 0x0018)
  // but move the main program code (ie 0x101A) 
  #option org_program = APP_PROGRAM_START + ($18 + 2)   // account for vector table size
#else
  // running under a bootloader
  // move everything... reset and int vectors go to 0x1000, 0x1008, and 0x1018
  // and main program after that (0x101A)
  #option org_reset = APP_PROGRAM_START
#endif

//
//-----------------------------------------------------------------------------
// SF startup initialization
//-----------------------------------------------------------------------------
// note: include startup.bas first before any other files as it has some 
// initialization that needs to be done right at startup
// this is VERY important if you're using a bootloader
//
// set WDT option to allow clrwdt() to tickle the wdt (since it may be enabled), 
// but do not change the config setting
//
#option WDT = true
include "startup.bas"

//
//-----------------------------------------------------------------------------
// config settings
// Note: The bootloader provides the ability to reprogram the config settings,
// so it's possible to build the bootloader and the application with different
// settings. However, this can be very dangerous and can result in non-functional
// operation, or even a system unable to boot. It's probably best to just set 
// them once here, otherwise use great caution.
//-----------------------------------------------------------------------------
//
// include the USER_INIT file specified above
include USER_INIT

// debugger support file included after user init in case we need to override any 
// config settings the debugger may need (ie WDT off)
#if isoption(MPLAB_ICD) and (MPLAB_ICD)
  include "mplabicd.bas"
#endif

// we need a clock setup, so might as well use the USB one
#option USB_CLOCK_USE_DELAYMS = true
include "USBclock.bas"

//
//-----------------------------------------------------------------------------
// IPC module (optional)
// used for application->bootloader communications
// If used, this file MUST be included before any other libraries/startup code
// that might modify the RCON register
//-----------------------------------------------------------------------------
//
#option USE_IPC = true
include "ipcBootloader.bas"

//
//-----------------------------------------------------------------------------
// options to test bootloader EEPROM and/or UserID functionality (optional)
// used to set data into the output HEX file
// note: these areas are not supported on J series devices
//-----------------------------------------------------------------------------
//
#option INCLUDE_USERID_DATA = true
#option INCLUDE_EEPROM_DATA = true

// this will put data into the hex file at 0x200000
#if (INCLUDE_USERID_DATA)
  // UserID init
  include "userid.bas"
#endif

// this will put data into the hex file at 0xF00000
#if (INCLUDE_EEPROM_DATA)
  // eeprom data
  eeprom = ($55,1,2,3,4,5,6,7,8,$AA)
  // include routine so that we can read it
  include "eeprom.bas"
#endif

// insert a large amount of const program data so we simulate having a large program
// it defaults to about 8K bytes, but you can modify the file to make it larger
// set it to 0 to not include the file and only use what's here in this module
#option HEX_DATA_BLOCK = 8*1024        // 8K-64K in 8K steps (0 to disable)
#if (HEX_DATA_BLOCK > 0)
  include "testprogram_data.bas"
#endif

//
//-----------------------------------------------------------------------------
// User hardware routines
// **** customize to match your hardware ****
//-----------------------------------------------------------------------------
// the userio module maps some of the hardware defined in the USER_INIT
// file to some friendly names, including UserInit()/Exit(), UserBootselect(),
// and UserLedOn()/Off()/Toggle()
include "userio.bas"

//
//-----------------------------------------------------------------------------
// interrupt handlers
//-----------------------------------------------------------------------------
// declare some interrupt routines so we can check that the vector locations
// get set properly. we don't need them to do anything as they're not really 
// going to be used

// intr priority level constants
public const
    ipLow = 1,
    ipHigh = 2
    
// vector 0x0008
interrupt isr_high(ipHigh)
end interrupt

// vector 0x0018
interrupt isr_low(ipLow)
end interrupt

//
//-----------------------------------------------------------------------------
// main application code
//-----------------------------------------------------------------------------
//
dim b as byte

#if (INCLUDE_USERID_DATA)
  dim id(USER_ID_SIZE) as byte      // buffer to hold userID locations
#endif

#if (INCLUDE_EEPROM_DATA)
  dim eeb(10) as byte               // buffer to hold eeprom data
#endif

main:
    IPC.reset_rcon()
    USBClockInit()
    userInit()
    'userClockCheck()
    
    b = 0
  
  #if (INCLUDE_EEPROM_DATA)
    // read eeprom data into our buffer
    for b = 0 to bound(eeb)
        eeb(b) = EE.ReadByte(b)
    next        
  #endif
  
  #if (INCLUDE_USERID_DATA)
    // read userID area into our buffer
    userID.Read(id)
  #endif
  
    // let SF think we're going to be using interrupts
    // this is just here so we get all the vector locations set
    // **NOTE: this assumes that all of the IE flags in the INTCONx and PIEx registers
    // are 0 so that nothing will actually happen when they get enabled**
    enable(isr_low)
    enable(isr_high)
  
    while (true)
        UserLedToggle()
        delayms(500)
        
        // check the boot select switch
        // if it's pressed, wait until it's released and then invoke the bootloader
        if (UserBootselect()) then
            repeat
                delayms(100)                // toggle LED rapidly until sw is released...
                UserLedToggle()
            until (UserBootselect() = false)
            UserExit()
          #if (USE_IPC)                     // set flag to let bootloader know we want it to run
            IPC.run_bootloader()
          #else
            IPC.reset()                     // just do a normal reset... we should end up back here
          #endif
        endif
    end while
    
end program
