//
//***************************************************************************
// Name    : USBSFR.bas                                                     *
// Author  : David John Barker - rev.                                       *
// Notice  : Copyright (c) 2007 Mecanique                                   *
//         : All Rights Reserved                                            *
// Date    : 06/01/2007                                                     *
// Version : 1.4.3                                                          *
//                                                                          *
// History : 1.0 Release 06/01/2007                                         *
//           1.4 Add definitions for all 18F series USB chips               *
//               Add '#define USB_ENDPOINT_REGISTERS'                       *
//               Add sub SetUSBClock() and '#option ENABLE_USB_CLOCK'       *
//           1.4.3 Add definitions for J94 and 45K50 family                 *
//              remove structure definitions for interrupts, and replace    *
//              with USB_IF, USB_IE, and USB_IP                             *
//              move usb clock setup into it's own module                   *
// Notes   : This module implements SFRs as structures, which allows you to *
//         : refer to bits using their respective datasheet names. For      *
//         : example, you can use something like UIRBits.STALLIF = 0. Note  *
//         : that if you want to access as a complete byte, you should use  *
//         : UIRBits._byte = $00 - this is because the compiler sees the    *
//         : variables as structures and not bytes                          *
//***************************************************************************
//
module USBSFR

// UFRML
structure TUFRML
    _byte as byte
    FRM0      as _byte.0
    FRM1      as _byte.1
    FRM2      as _byte.2
    FRM3      as _byte.3
    FRM4      as _byte.4
    FRM5      as _byte.5
    FRM6      as _byte.6
    FRM7      as _byte.7
end structure

// UFRMH
structure TUFRMH
    _byte as byte
    FRM8      as _byte.0
    FRM9      as _byte.1
    FRM10     as _byte.2
end structure

// UIR
structure TUIR
    _byte as byte
    URSTIF    as _byte.0
    UERRIF    as _byte.1
    ACTVIF    as _byte.2
    TRNIF     as _byte.3
    IDLEIF    as _byte.4
    STALLIF   as _byte.5
    SOFIF     as _byte.6
end structure

// UIE
structure TUIE
    _byte as byte
    URSTIE    as _byte.0
    UERRIE    as _byte.1
    ACTVIE    as _byte.2
    TRNIE     as _byte.3
    IDLEIE    as _byte.4
    STALLIE   as _byte.5
    SOFIE     as _byte.6
end structure
// UIE bit constants
public const
    UIE_ENABLE_ALL = $7F,      // all interrupts
    UIE_URSTIE = (1<<0),
    UIE_UERRIE = (1<<1),
    UIE_ACTVIE = (1<<2),
    UIE_TRNIE  = (1<<3),
    UIE_IDLEIE = (1<<4),
    UIE_STALLIE= (1<<5),
    UIE_SOFIE  = (1<<6)

// UEIR
structure TUEIR
    _byte as byte
    PIDEF     as _byte.0
    CRC5EF    as _byte.1
    CRC16EF   as _byte.2
    DFN8EF    as _byte.3
    BTOEF     as _byte.4
    BTSEF     as _byte.7
end structure

// UEIE
structure TUEIE
    _byte as byte
    PIDEE     as _byte.0
    CRC5EE    as _byte.1
    CRC16EE   as _byte.2
    DFN8EE    as _byte.3
    BTOEE     as _byte.4
    BTSEE     as _byte.7
end structure
// UIEI bit constants
public const
    UEIE_ENABLE_ALL = $9F      // all error interrupts

// USTAT
structure TUSTAT
    _byte as byte
    PPBI      as _byte.1
    DIR       as _byte.2
    ENDP0     as _byte.3
    ENDP1     as _byte.4
    ENDP2     as _byte.5
    ENDP3     as _byte.6      // unused in the 13K/14K50 (only 8 endpoints)
end structure

// UCON
structure TUCON
    _byte as byte
    SUSPND    as _byte.1
    RESUME    as _byte.2
    USBEN     as _byte.3
    PKTDIS    as _byte.4
    SE0       as _byte.5
    PPBRST    as _byte.6
end structure

// UADDR
structure TUADDR
    _byte as byte
    ADDR0     as _byte.0
    ADDR1     as _byte.1
    ADDR2     as _byte.2
    ADDR3     as _byte.3
    ADDR4     as _byte.4
    ADDR5     as _byte.5
    ADDR6     as _byte.6
end structure

// UCFG
structure TUCFG
    _byte as byte
    PPB0      as _byte.0
    PPB1      as _byte.1
    FSEN      as _byte.2
    UTRDIS    as _byte.3      // unimplemented on the 13K/14K50
    UPUEN     as _byte.4
    UOEMON    as _byte.6      // unimplemented on the 13K/14K50, 87J50
    UTEYE     as _byte.7
end structure

// UEP
structure TUEP
    _byte as byte
    EPSTALL   as _byte.0
    EPINEN    as _byte.1
    EPOUTEN   as _byte.2
    EPCONDIS  as _byte.3
    EPHSHK    as _byte.4
end structure

// INTCON
structure TINTCON
    _byte as byte
    RBIF      as _byte.0
    INT0IF    as _byte.1
    TMR0IF    as _byte.2
    RBIE      as _byte.3
    INT0IE    as _byte.4
    TMR0IE    as _byte.5

    PEIE      as _byte.6
    GIEL      as _byte.6
    GIE       as _byte.7
    GIEH      as _byte.7

  #if _device in (18F13K50, 18F14K50)
    RABIF     as _byte.0
    RABIE     as _byte.3
  #endif
end structure

//
//------------------------------------------------------------------------
// PIC18 INTCON register structure
//------------------------------------------------------------------------
//
public dim
    INTCONBits  as TINTCON absolute $0FF2

//
//------------------------------------------------------------------------
// USB register and interrupt bit definitions
//------------------------------------------------------------------------
//
#if _device in (18F2450, 18F4450)
#define USB_ENDPOINT_REGISTERS = 16     // number of EP registers

public dim 
    USB_IF as PIR2.bits(5),
    USB_IE as PIE2.bits(5),
    USB_IP as IPR2.bits(5)

public dim
   UFRMLBits    as TUFRML absolute $0F66,
   UFRMHBits    as TUFRMH absolute $0F67,
   UIRBits      as TUIR   absolute $0F68,
   UIEBits      as TUIE   absolute $0F69,
   UEIRBits     as TUEIR  absolute $0F6A,
   UEIEBits     as TUEIE  absolute $0F6B,
   USTATBits    as TUSTAT absolute $0F6C,
   UCONBits     as TUCON  absolute $0F6D,
   UADDRBits    as TUADDR absolute $0F6E,
   UCFGBits     as TUCFG  absolute $0F6F,

   UEP0Bits     as TUEP   absolute $0F70,
   UEP1Bits     as TUEP   absolute $0F71,
   UEP2Bits     as TUEP   absolute $0F72,
   UEP3Bits     as TUEP   absolute $0F73,
   UEP4Bits     as TUEP   absolute $0F74,
   UEP5Bits     as TUEP   absolute $0F75,
   UEP6Bits     as TUEP   absolute $0F76,
   UEP7Bits     as TUEP   absolute $0F77,
   UEP8Bits     as TUEP   absolute $0F78,
   UEP9Bits     as TUEP   absolute $0F79,
   UEP10Bits    as TUEP   absolute $0F7A,
   UEP11Bits    as TUEP   absolute $0F7B,
   UEP12Bits    as TUEP   absolute $0F7C,
   UEP13Bits    as TUEP   absolute $0F7D,
   UEP14Bits    as TUEP   absolute $0F7E,
   UEP15Bits    as TUEP   absolute $0F7F

#elseif _device in (18F2455, 18F2550, 18F4455, 18F4550, 18F2458, 18F2553, 18F4458, 18F4553)
#define USB_ENDPOINT_REGISTERS = 16     // number of EP registers

public dim 
    USB_IF as PIR2.bits(5),
    USB_IE as PIE2.bits(5),
    USB_IP as IPR2.bits(5)

public dim
   UFRMLBits    as TUFRML absolute $0F66,
   UFRMHBits    as TUFRMH absolute $0F67,
   UIRBits      as TUIR   absolute $0F68,
   UIEBits      as TUIE   absolute $0F69,
   UEIRBits     as TUEIR  absolute $0F6A,
   UEIEBits     as TUEIE  absolute $0F6B,
   USTATBits    as TUSTAT absolute $0F6C,
   UCONBits     as TUCON  absolute $0F6D,
   UADDRBits    as TUADDR absolute $0F6E,
   UCFGBits     as TUCFG  absolute $0F6F,

   UEP0Bits     as TUEP   absolute $0F70,
   UEP1Bits     as TUEP   absolute $0F71,
   UEP2Bits     as TUEP   absolute $0F72,
   UEP3Bits     as TUEP   absolute $0F73,
   UEP4Bits     as TUEP   absolute $0F74,
   UEP5Bits     as TUEP   absolute $0F75,
   UEP6Bits     as TUEP   absolute $0F76,
   UEP7Bits     as TUEP   absolute $0F77,
   UEP8Bits     as TUEP   absolute $0F78,
   UEP9Bits     as TUEP   absolute $0F79,
   UEP10Bits    as TUEP   absolute $0F7A,
   UEP11Bits    as TUEP   absolute $0F7B,
   UEP12Bits    as TUEP   absolute $0F7C,
   UEP13Bits    as TUEP   absolute $0F7D,
   UEP14Bits    as TUEP   absolute $0F7E,
   UEP15Bits    as TUEP   absolute $0F7F

#elseif _device in (18F24J50, 18F25J50, 18F26J50, 18F44J50, 18F45J50, 18F46J50)
#define USB_ENDPOINT_REGISTERS = 16     // number of EP registers

public dim 
    USB_IF as PIR2.bits(4),
    USB_IE as PIE2.bits(4),
    USB_IP as IPR2.bits(4)

public dim
   //ACCESS BANK SPECIAL FUNCTION REGISTER MAP
   UFRMLBits    as TUFRML absolute $0F60,
   UFRMHBits    as TUFRMH absolute $0F61,
   UIRBits      as TUIR   absolute $0F62,
   UEIRBits     as TUEIR  absolute $0F63,
   USTATBits    as TUSTAT absolute $0F64,
   UCONBits     as TUCON  absolute $0F65,
   //NON-ACCESS BANK SPECIAL FUNCTION REGISTER MAP
   UIEBits      as TUIE   absolute $0F36,
   UEIEBits     as TUEIE  absolute $0F37,
   UADDRBits    as TUADDR absolute $0F38,
   UCFGBits     as TUCFG  absolute $0F39,

   UEP0Bits     as TUEP   absolute $0F26,
   UEP1Bits     as TUEP   absolute $0F27,
   UEP2Bits     as TUEP   absolute $0F28,
   UEP3Bits     as TUEP   absolute $0F29,
   UEP4Bits     as TUEP   absolute $0F2A,
   UEP5Bits     as TUEP   absolute $0F2B,
   UEP6Bits     as TUEP   absolute $0F2C,
   UEP7Bits     as TUEP   absolute $0F2D,
   UEP8Bits     as TUEP   absolute $0F2E,
   UEP9Bits     as TUEP   absolute $0F2F,
   UEP10Bits    as TUEP   absolute $0F30,
   UEP11Bits    as TUEP   absolute $0F31,
   UEP12Bits    as TUEP   absolute $0F32,
   UEP13Bits    as TUEP   absolute $0F33,
   UEP14Bits    as TUEP   absolute $0F34,
   UEP15Bits    as TUEP   absolute $0F35

#elseif _device in (18F26J53, 18F27J53, 18F46J53, 18F47J53)
#define USB_ENDPOINT_REGISTERS = 16     // number of EP registers

public dim 
    USB_IF as PIR2.bits(4),
    USB_IE as PIE2.bits(4),
    USB_IP as IPR2.bits(4)

public dim
   //ACCESS BANK SPECIAL FUNCTION REGISTER MAP
   UFRMLBits    as TUFRML absolute $0F60,
   UFRMHBits    as TUFRMH absolute $0F61,
   UIRBits      as TUIR   absolute $0F62,
   UEIRBits     as TUEIR  absolute $0F63,
   USTATBits    as TUSTAT absolute $0F64,
   UCONBits     as TUCON  absolute $0F65,
   //NON-ACCESS BANK SPECIAL FUNCTION REGISTER MAP
   UIEBits      as TUIE   absolute $0F36,
   UEIEBits     as TUEIE  absolute $0F37,
   UADDRBits    as TUADDR absolute $0F38,
   UCFGBits     as TUCFG  absolute $0F39,

   UEP0Bits     as TUEP   absolute $0F26,
   UEP1Bits     as TUEP   absolute $0F27,
   UEP2Bits     as TUEP   absolute $0F28,
   UEP3Bits     as TUEP   absolute $0F29,
   UEP4Bits     as TUEP   absolute $0F2A,
   UEP5Bits     as TUEP   absolute $0F2B,
   UEP6Bits     as TUEP   absolute $0F2C,
   UEP7Bits     as TUEP   absolute $0F2D,
   UEP8Bits     as TUEP   absolute $0F2E,
   UEP9Bits     as TUEP   absolute $0F2F,
   UEP10Bits    as TUEP   absolute $0F30,
   UEP11Bits    as TUEP   absolute $0F31,
   UEP12Bits    as TUEP   absolute $0F32,
   UEP13Bits    as TUEP   absolute $0F33,
   UEP14Bits    as TUEP   absolute $0F34,
   UEP15Bits    as TUEP   absolute $0F35

#elseif _device in (18F65J50, 18F66J50, 18F66J55, 18F67J50, 18F85J50, 18F86J50, 18F86J55, 18F87J50)
#define USB_ENDPOINT_REGISTERS = 16     // number of EP registers

public dim 
    USB_IF as PIR2.bits(4),
    USB_IE as PIE2.bits(4),
    USB_IP as IPR2.bits(4)

public dim
   UFRMLBits    as TUFRML absolute $0F60,
   UFRMHBits    as TUFRMH absolute $0F61,
   UIRBits      as TUIR   absolute $0F62,
   UEIRBits     as TUEIR  absolute $0F63,
   USTATBits    as TUSTAT absolute $0F64,
   UCONBits     as TUCON  absolute $0F65,
   UIEBits      as TUIE   absolute $0F5C,
   UEIEBits     as TUEIE  absolute $0F5D,
   UADDRBits    as TUADDR absolute $0F5E,
   UCFGBits     as TUCFG  absolute $0F5F,

   UEP0Bits     as TUEP   absolute $0F4C,
   UEP1Bits     as TUEP   absolute $0F4D,
   UEP2Bits     as TUEP   absolute $0F4E,
   UEP3Bits     as TUEP   absolute $0F4F,
   UEP4Bits     as TUEP   absolute $0F50,
   UEP5Bits     as TUEP   absolute $0F51,
   UEP6Bits     as TUEP   absolute $0F52,
   UEP7Bits     as TUEP   absolute $0F53,
   UEP8Bits     as TUEP   absolute $0F54,
   UEP9Bits     as TUEP   absolute $0F55,
   UEP10Bits    as TUEP   absolute $0F56,
   UEP11Bits    as TUEP   absolute $0F57,
   UEP12Bits    as TUEP   absolute $0F58,
   UEP13Bits    as TUEP   absolute $0F59,
   UEP14Bits    as TUEP   absolute $0F5A,
   UEP15Bits    as TUEP   absolute $0F5B

#elseif _device in (18F65J94, 18F66J94, 18F67J94, 18F85J94, 18F86J94, 18F87J94, 18F95J94, 18F96J94, 18F97J94) or _
        _device in (18F66J99, 18F86J99, 18F96J99)
#define USB_ENDPOINT_REGISTERS = 16     // number of EP registers

public dim 
    USB_IF as PIR2.bits(4),
    USB_IE as PIE2.bits(4),
    USB_IP as IPR2.bits(4)

public dim
   UFRMLBits    as TUFRML absolute $0F7A,
   UFRMHBits    as TUFRMH absolute $0F79,
   UIRBits      as TUIR   absolute $0F78,
   UEIRBits     as TUEIR  absolute $0F77,
   USTATBits    as TUSTAT absolute $0F76,
   UCONBits     as TUCON  absolute $0F75,
   UIEBits      as TUIE   absolute $0E10,
   UEIEBits     as TUEIE  absolute $0E0F,
   UADDRBits    as TUADDR absolute $0F74,
   UCFGBits     as TUCFG  absolute $0E11,

   UEP0Bits     as TUEP   absolute $0DFF,
   UEP1Bits     as TUEP   absolute $0E00,
   UEP2Bits     as TUEP   absolute $0E01,
   UEP3Bits     as TUEP   absolute $0E02,
   UEP4Bits     as TUEP   absolute $0E03,
   UEP5Bits     as TUEP   absolute $0E04,
   UEP6Bits     as TUEP   absolute $0E05,
   UEP7Bits     as TUEP   absolute $0E06,
   UEP8Bits     as TUEP   absolute $0E07,
   UEP9Bits     as TUEP   absolute $0E08,
   UEP10Bits    as TUEP   absolute $0E09,
   UEP11Bits    as TUEP   absolute $0E0A,
   UEP12Bits    as TUEP   absolute $0E0B,
   UEP13Bits    as TUEP   absolute $0E0C,
   UEP14Bits    as TUEP   absolute $0E0D,
   UEP15Bits    as TUEP   absolute $0E0E

#elseif _device in (18F13K50, 18F14K50)
#define USB_ENDPOINT_REGISTERS = 8     // number of EP registers

public dim 
    USB_IF as PIR2.bits(2),
    USB_IE as PIE2.bits(2),
    USB_IP as IPR2.bits(2)

public dim
   UEIEBits     as TUEIE  absolute $0F5B,
   UADDRBits    as TUADDR absolute $0F5C,
   UFRMLBits    as TUFRML absolute $0F5D,
   UFRMHBits    as TUFRMH absolute $0F5E,
   UEIRBits     as TUEIR  absolute $0F5F,
   UIEBits      as TUIE   absolute $0F60,
   UCFGBits     as TUCFG  absolute $0F61,
   UIRBits      as TUIR   absolute $0F62,
   USTATBits    as TUSTAT absolute $0F63,
   UCONBits     as TUCON  absolute $0F64,

   UEP0Bits     as TUEP   absolute $0F53,
   UEP1Bits     as TUEP   absolute $0F54,
   UEP2Bits     as TUEP   absolute $0F55,
   UEP3Bits     as TUEP   absolute $0F56,
   UEP4Bits     as TUEP   absolute $0F57,
   UEP5Bits     as TUEP   absolute $0F58,
   UEP6Bits     as TUEP   absolute $0F59,
   UEP7Bits     as TUEP   absolute $0F5A

#elseif _device in (18F24K50, 18F25K50, 18F45K50)
#define USB_ENDPOINT_REGISTERS = 16     // number of EP registers

public dim 
    USB_IF as PIR3.bits(2),
    USB_IE as PIE3.bits(2),
    USB_IP as IPR3.bits(2)

public dim
   UFRMLBits    as TUFRML absolute $0F68,
   UFRMHBits    as TUFRMH absolute $0F69,
   UIRBits      as TUIR   absolute $0F65,
   UEIRBits     as TUEIR  absolute $0F67,
   USTATBits    as TUSTAT absolute $0F61,
   UCONBits     as TUCON  absolute $0F60,
   UIEBits      as TUIE   absolute $0F64,
   UEIEBits     as TUEIE  absolute $0F66,
   UADDRBits    as TUADDR absolute $0F63,
   UCFGBits     as TUCFG  absolute $0F62,

   UEP0Bits     as TUEP   absolute $0F6A,
   UEP1Bits     as TUEP   absolute $0F6B,
   UEP2Bits     as TUEP   absolute $0F6C,
   UEP3Bits     as TUEP   absolute $0F6D,
   UEP4Bits     as TUEP   absolute $0F6E,
   UEP5Bits     as TUEP   absolute $0F6F,
   UEP6Bits     as TUEP   absolute $0F70,
   UEP7Bits     as TUEP   absolute $0F71,
   UEP8Bits     as TUEP   absolute $0F72,
   UEP9Bits     as TUEP   absolute $0F73,
   UEP10Bits    as TUEP   absolute $0F74,
   UEP11Bits    as TUEP   absolute $0F75,
   UEP12Bits    as TUEP   absolute $0F76,
   UEP13Bits    as TUEP   absolute $0F77,
   UEP14Bits    as TUEP   absolute $0F78,
   UEP15Bits    as TUEP   absolute $0F79

#else
  // add new device USB register and interrupt definitions here
  #error "unsupported USB device"
#endif

end module
