//
//---------------------------------------------------------------------------
// Name    : ipc.bas
// Author  : Jerry Messina
// Notice  : Copyright (c) 2011
//         : All Rights Reserved
// Date    : 4/17/2011
// Version : 1.0
// Notes   : InterProcessCommunications (IPC) module
//         : application code -> bootloader
//---------------------------------------------------------------------------
//
module ipc

//
// This module outlines a method for the application to communicate to the
// bootloader, allowing the app to run the bootloader without requiring an
// IO pin or switch press.
//
// Define the 16-bit register used for application->bootloader communications.
// The bootloader will check this location if RCON indicates that a reset
// event has been caused by a software RESET instruction. If it contains the
// data "BL" ($424C), then the bootloader will remain active and not run the
// application. To run the bootloader simply set the register contents to
// $424C and generate a software RESET.
//
// There are a number of "16-bit" PIC18 registers that are unchanged by a software
// RESET instruction, including the PROD (PRODH/PRODL), the TMR1 (TMR1H/TMR1L),
// and the ADRES (ADRESH/ADRESL) registers. The PROD registers are the easiest,
// since they are present in all PIC18 devices and require no special treatment
// (ie making sure timer is stopped, no automatic AD conversion running, etc)
// which would complicate things. You are not restricted to using pic SFRs...
// you could also use two bytes of RAM (perhaps the top two) as long as they're
// at a known, fixed address and are not changed by the SF startup code. The only
// requirement is that the registers remain unchanged until this module gets a
// chance to read them.
//
// To invoke the bootloader from the application program, execute the following:
//      INTCON = 0      'disable interrupts
//      IPC_REG = REQUEST_BOOTLOADER
//      reset()         ' where reset() generates a PIC18 software RESET
//
// To generate a normal reset (and skip the bootloader), clear the IPC_REG first:
//      INTCON = 0      'disable interrupts
//      IPC_REG = $0000
//      reset()         ' where reset() generates a PIC18 software RESET
//
// This feature can be controlled using the '#option USE_IPC' setting.
// Set '#option USE_IPC = false' to disable this feature.
//
// Note that this simple method is by no means foolproof. If it is enabled in the
// bootloader and the application does not set IPC_REG before issuing a RESET,
// then there is a (small) chance that the bootloader will misinterpret the reset
// and remain active. You can use the routines below in the application to help avoid
// this.
//
// This is placed in its own module so that the registers can be read upon program
// entry before any other code executes. It should be included first before any
// other modules (at least for the bootloader).
//

// module enabled by default
#option USE_IPC = true

// constant "BL"
public const REQUEST_BOOTLOADER as word = $424C

// define the pic registers used for IPC. here we use PRODH/PRODL
public dim IPC_REG as PRODL.AsWord

// RCON - reset status reg (note: status bits are low-active)
public structure RCON_REG_T
    b as byte
    BOR     as b.0
    POR     as b.1
    PD      as b.2
    _TO     as b.3
    RI      as b.4
    IPEN    as b.7
end structure

//
// these routines are for use by the application code
//

// set all of the RCON reset status bits to 1 (leave IPEN alone)
// you should do this upon entry to your main program
public inline sub reset_rcon()
    RCON = RCON or %00011111
end sub

// run_bootloader - invoke the bootloader via software
public inline sub run_bootloader()
    INTCON = 0                      // disable interrupts
    IPC_REG = REQUEST_BOOTLOADER    // set reg to "BL"
    asm
      reset                         // reboot using software reset
    end asm
end sub

// perform normal reset/reboot
public inline sub reset()
    INTCON = 0                      // disable interrupts
    IPC_REG = $0000                 // skip bootloader... just reboot
    asm
      reset                         // reboot using software reset
    end asm
end sub

//
// module initialization
//
#if (USE_IPC)
public dim ipc_request as word          // shadow copy of IPC_REG
public dim _rcon as RCON_REG_T          // shadow copy of RCON register

// read the register used for IPC
ipc_request = IPC_REG

// read the RCON register before anyone has a chance to change it
_rcon.b = RCON
#endif      // USE_IPC

end module
