//
//-----------------------------------------------------------------------------
// Name    : main_cdc.bas
// Author  : Jerry Messina
// Date    : 1/1/2015
// Version : 1.4.3
// Notes   : USB CDC main application example
//         :
// V1.4.3  : added support for J94 and K50 families
//         : add SF _protected region support (removed EXTENDED_RAM)
//-----------------------------------------------------------------------------
//
program main_CDC

//
//-----------------------------------------------------------------------------
// hardware platforms (uncomment one section/customize for your hardware)
// the USER_INIT file for each platform contains the pic config settings, 
// userInit() routine, LED definitions, and any misc setup info
//-----------------------------------------------------------------------------
//
'{
// Sparkfun USB Bit Whacker v1.2 18F2553 DEV-00762 (www.sparkfun.com)
device = 18F2553        // 18F2550 family
clock = 48
#option USER_INIT = "UBW_2553.bas"
'}

{
// Microchip LPC (LowPinCount) USB dev kit 18F14K50 DM164127
device = 18F14K50
clock = 48
#option USER_INIT = "LPC_14K50.bas"
}

{
// Microchip USB PIM 18F47J53 MA180029
device = 18F47J53
clock = 48
#option USER_INIT = "PIM_47J53.bas"
}

{
// Microchip USB PIM 18F87J94 MA180033
device = 18F87J94
clock = 64          // clock setup supports 64MHz or 32MHz
#option USER_INIT = "PIM_87J94.bas"
}

{
// Microchip PICDEM FS USB dev board 18F45K50 DM163025-1
device = 18F45K50
clock = 48
#option USER_INIT = "PICDEM_FS_45K50.bas"
}


//
//-----------------------------------------------------------------------------
// MPLAB ICD debugger support (true/false)
//-----------------------------------------------------------------------------
//
#option MPLAB_ICD = false

//
//-----------------------------------------------------------------------------
// hardware setup
//-----------------------------------------------------------------------------
//
// include the USER_INIT file specified above
include USER_INIT

// debugger support file included after user init in case we need to override any 
// config settings the debugger may need (ie WDT off)
#if isoption(MPLAB_ICD) and (MPLAB_ICD)
  include "mplabicd.bas"
#endif

//
//-----------------------------------------------------------------------------
// USB options and includes
//-----------------------------------------------------------------------------
//
// enable display of build status messages to the IDE Results window
#option SHOW_USB_BUILD_MSGS = false      // false (off) by default

// interrupt/polled mode (true=interrupt, false=polled)
// note: this option defaults to true in USBConfig
#option USB_SERVICE = true

// include custom USB clock initialization routine
// (optional... will use default in USBClock.bas if none specified)
'#option USER_USB_CLOCK_INIT = "userUsbClock.bas"

// these options control the amount of ram used by CDC
// normally you don't need to specify these, and can exclude them (or leave them
// set to the default values shown)
// if you're trying to get the minimal memory footprint then reduce these
//
// enable TX double buffering (uses 2 CDC_TXB_SIZE tx user buffers)
#option CDC_DOUBLE_BUFFER_TX = true
// cdc bulk in/bulk out endpoint buffer sizes (max 64 bytes for full-speed USB)
#option CDC_TX_SIZE = 64
#option CDC_RX_SIZE = 64
// cdc rx and tx user buffer sizes
#option CDC_TXB_SIZE = 64           // defaults to endpoint CDC_TX_SIZE
#option CDC_RXB_SIZE = 64           // defaults to endpoint CDC_RX_SIZE

// import USB modules...
include "usbcdc.bas"
include "usbsystem.bas"

//
//-----------------------------------------------------------------------------
// main program example selection
// select one of the following examples by setting 
// #option DEMO_EXAMPLE = 1, 2, 3, or 4
//-----------------------------------------------------------------------------
//
#option DEMO_EXAMPLE = 1

//
//-------------------------------------------------------------
// example 1
// read bytes and echo them back 
// use SF SerialCommunicator or another terminal program such as
// TeraTerm (http://en.sourceforge.jp/projects/ttssh2/)
//-------------------------------------------------------------
//
#if (DEMO_EXAMPLE = 1)
userInit()      // hdw specific pic init

// wait to connect to USB...
low(LED1)
repeat
    USBservice()        // if using interrupts, calls to USBService() will be ignored
until (CDC.Attached)
high(LED1)

while (true)
    if (CDC.Attached) then
        high(LED2)
    else
        low(LED2)
    endif

    if (CDC.DataAvailable()) then
        high(LED1)            // flash led to indicate we've got a packet
        CDC.WriteByte(CDC.ReadByte())   // read and echo data
        low(LED1)
    endif

    // if the USB_SERVICE option has been disabled, you MUST poll the USB connection
    // every 1ms or so to keep it alive and handle pending transactions on the bus...
    USBservice()
end while
#endif      // example 1


//
//-------------------------------------------------------------
// example 2
// use ReadArray/WriteArray to read multiple bytes
//-------------------------------------------------------------
//
#if (DEMO_EXAMPLE = 2)
dim array(6) as byte        // max of 6 chars!

userInit()      // hdw specific pic init

// wait to connect to USB...
low(LED1)
repeat
    USBservice()        // if using interrupts, calls to USBService() will be ignored
until (CDC.Attached)
high(LED1)

while (true)
    if (CDC.Attached) then
        high(LED2)
    else
        low(LED2)
    endif

    CDC.Write("send 6 bytes> ")         // 6 = sizeof(array)
    CDC.ReadArray(array)                // waits for sizeof(array) number of bytes
    high(LED1)
    CDC.WriteArray(array)
    CDC.WriteByte(13)
    CDC.WriteByte(10)
    low(LED1)

    // if the USB_SERVICE option has been disabled, you MUST poll the USB connection
    // every 1ms or so to keep it alive and handle pending transactions on the bus...
    USBservice()
end while
#endif      // example 2


//
//-------------------------------------------------------------
// example 3
// use string terminator to read an input string
//-------------------------------------------------------------
//
#if (DEMO_EXAMPLE = 3)
// the max size of an undimensioned string is 24 chars (23+'\0')
// if you want to accept larger strings then you should dimension the string as appropriate
// here we set it to be the size of the input buffer
dim instr as string(CDC_RXB_SIZE+1)

userInit()      // hdw specific pic init

// wait to connect to USB...
low(LED1)
repeat
    USBservice()        // if using interrupts, calls to USBService() will be ignored
until (CDC.Attached)
high(LED1)

CDC.ReadTerminator = 13         // set CR as input terminator
while (true)
    if (CDC.Attached) then
        high(LED2)
    else
        low(LED2)
    endif

    CDC.Write("send string + CR> ")
    CDC.Read(instr)
    high(LED1)
    CDC.Write(instr)
    CDC.WriteByte(13)
    CDC.WriteByte(10)
    low(LED1)

    // if the USB_SERVICE option has been disabled, you MUST poll the USB connection
    // every 1ms or so to keep it alive and handle pending transactions on the bus...
    USBservice()
end while
#endif      // example 3

//
//-------------------------------------------------------------
// example 4
// serial BREAK handler event
//-------------------------------------------------------------
//
#if (DEMO_EXAMPLE = 4)
// flash led a few times when a comm SEND_BREAK occurs
event break_handler()
    dim b as byte
    for b = 0 to 5
        toggle(LED2)
        delayms(100)    // long delay here isn't a good idea... just for demo
    next        
end event

userInit()      // hdw specific pic init

// wait to connect to USB...
low(LED1)
repeat
    USBservice()        // if using interrupts, calls to USBService() will be ignored
until (CDC.Attached)
high(LED1)

// set break event handler
// SEND_BREAK is processed by USBSystem, and will call the user-defined event (if there is one)
CDC.SendBreakEvent = break_handler

while (true)
    if (CDC.Attached) then
        high(LED2)
    else
        low(LED2)
    endif

    // if BREAK detected, the LED2 should toggle
    // otherwise, we'll just echo chars
    if (CDC.DataAvailable()) then
        high(LED1)            // flash led to indicate we've got a packet
        CDC.WriteByte(CDC.ReadByte())   // read and echo data
        low(LED1)
    endif

    // if the USB_SERVICE option has been disabled, you MUST poll the USB connection
    // every 1ms or so to keep it alive and handle pending transactions on the bus...
    USBservice()
end while
#endif      // example 4

//
//-------------------------------------------------------------
// example 5
// USB self-powered device
// read bytes and echo them back 
//-------------------------------------------------------------
//
#if (DEMO_EXAMPLE = 5)
userInit()      // hdw specific pic init

// wait to connect to USB...
low(LED1)

// if we're self powered and can detect USB VBUS then wait until it's sensed high (usb cable attached)
#if ((USB_SELF_POWER = 1) and USB_BUS_SENSE_IO)
  while (USB_BUS_SENSE_IN = 0)
  end while
  USBDeviceAttach()
#endif
// wait until we're enumerated
repeat
    USBservice()        // if using interrupts, calls to USBService() will be ignored
until (CDC.Attached)
// we're connected...
high(LED1)

while (true)
    if (CDC.Attached) then
        high(LED2)
    else
        low(LED2)
    endif

    if (CDC.DataAvailable()) then
        high(LED1)            // flash led to indicate we've got a packet
        CDC.WriteByte(CDC.ReadByte())   // read and echo data
        low(LED1)
    endif

    // if the USB_SERVICE option has been disabled, you MUST poll the USB connection
    // every 1ms or so to keep it alive and handle pending transactions on the bus...
    USBservice()
end while
#endif      // example 5

end program

