//
//-----------------------------------------------------------------------------
// Name    : main_hid.bas
// Author  : Jerry Messina
// Date    : 1/1/2015
// Version : 1.4.3
// Notes   : USB HID main application example
//         :
// V1.4.3  : added support for J94 and K50 families
//         : add SF _protected region support (removed EXTENDED_RAM)
//         : add custom HID Reports file support
//         : add custom usb clock setup support
//-----------------------------------------------------------------------------
//
program main_HID

//
//-----------------------------------------------------------------------------
// hardware platforms (uncomment one section/customize for your hardware)
// the USER_INIT file for each platform contains the pic config settings, 
// userInit() routine, LED definitions, and any misc setup info
//-----------------------------------------------------------------------------
//
'{
// Sparkfun USB Bit Whacker v1.2 18F2553 DEV-00762 (www.sparkfun.com)
device = 18F2553        // 18F2550 family
clock = 48
#option USER_INIT = "UBW_2553.bas"
'}

{
// Microchip LPC (LowPinCount) USB dev kit 18F14K50 DM164127
device = 18F14K50
clock = 48
#option USER_INIT = "LPC_14K50.bas"
}

{
// Microchip USB PIM 18F47J53 MA180029
device = 18F47J53
clock = 48
#option USER_INIT = "PIM_47J53.bas"
}

{
// Microchip USB PIM 18F87J94 MA180033
device = 18F87J94
clock = 64          // clock setup supports 64MHz or 32MHz
#option USER_INIT = "PIM_87J94.bas"
}

{
// Microchip PICDEM FS USB dev board 18F45K50 DM163025-1
device = 18F45K50
clock = 48
#option USER_INIT = "PICDEM_FS_45K50.bas"
}

//
//-----------------------------------------------------------------------------
// MPLAB ICD debugger support (true/false)
//-----------------------------------------------------------------------------
//
#option MPLAB_ICD = false

//
//-----------------------------------------------------------------------------
// hardware setup
//-----------------------------------------------------------------------------
//
// include the USER_INIT file specified above
include USER_INIT

// debugger support file included after user init in case we need to override any 
// config settings the debugger may need (ie WDT off)
#if isoption(MPLAB_ICD) and (MPLAB_ICD)
  include "mplabicd.bas"
#endif

//
//-----------------------------------------------------------------------------
// main program example selection
// select one of the following examples by setting 
//   #option DEMO_EXAMPLE = 1, 2, or 3
// note: as you enable the different examples, the USER_CUSTOM_HID_REPORTS
// option will be changed to include the custom HID reports file (if used)
//-----------------------------------------------------------------------------
//
#option DEMO_EXAMPLE = 1

//
//-----------------------------------------------------------------------------
// USB options and includes
//-----------------------------------------------------------------------------
//
// enable display of build status messages to the IDE Results window
#option SHOW_USB_BUILD_MSGS = false      // false (off) by default

// interrupt/polled mode (true=interrupt, false=polled)
// note: this option defaults to true in USBConfig
#option USB_SERVICE = true

// include custom USB clock initialization routine
// (optional... will use default in USBClock.bas if none specified)
'#option USER_USB_CLOCK_INIT = "userUsbClock.bas"

//
//-----------------------------------------------------------------------------
// optional custom HID report structures for ReadReport()/WriteReport()
//-----------------------------------------------------------------------------
// if you want to use custom report structures see the instructions
// in USBHID.bas for creating your own report definitions file, 
// and set '#option USER_CUSTOM_HID_REPORTS' to the name of your 
// custom report file. otherwise, leave the option undefined.
// example files can be found in the UserHIDReports directory
//
// here we change the filename based on the selected demo example
#if (DEMO_EXAMPLE = 1)
  #option USER_CUSTOM_HID_REPORTS = "UserReportsExample1.bas"
#elseif (DEMO_EXAMPLE = 2)
  #option USER_CUSTOM_HID_REPORTS = "UserReportsExample2.bas"
#else
    // no custom reports file req'd
#endif

// if a custom reports file has been specified, then include it here
// so that we can see the structure declarations (the actual definition
// is done in the usbhid module). this is only needed for mplab debugging,
// otherwise you get "Unsupported Struct" when viewing the TXReport and
// RxReport symbols in the watch window
#if isoption(USER_CUSTOM_HID_REPORTS) and MPLAB_ICD
  include USER_CUSTOM_HID_REPORTS
#endif

// import USB modules...
include "usbhid.bas"

// main program

//
//-----------------------------------------------------------------------------
// example 1 - generic text reports
// use the tool MikroE USB HIDTerminal (http://www.mikroe.com/upld/hidterminal.zip)
// set #option USER_CUSTOM_HID_REPORTS = "UserReportsExample1.bas"
//-----------------------------------------------------------------------------
//
#if (DEMO_EXAMPLE = 1)
userInit()      // hdw specific pic init

// wait to connect to USB...
low(LED1)
repeat
    USBservice()        // if using interrupts, calls to USBService() will be ignored
until (HID.Attached)
high(LED1)

// main program loop
// read report messages and send/echo response
while (true)
    if (HID.Attached) then
        high(LED2)
    else
        low(LED2)
    endif

    if (HID.DataAvailable()) then
        high(LED1)            // flash led to indicate we've got a packet
        HID.ReadReport()                             // read incoming report
        TXReport.Message = ">" + RXReport.Message    // just echo the received data
        HID.WriteReport()
        TXReport.Message = #13 + #10                 // along with CRLF
        HID.WriteReport()
        low(LED1)
    endif

    // if the USB_SERVICE option has been disabled, you MUST poll the USB connection
    // every 1ms or so to keep it alive and handle pending transactions on the bus...
    USBservice()
end while
#endif      // example 1

//
//-----------------------------------------------------------------------------
// example 2 - SF HIDExample tool
// use the tool http://www.sfcompiler.co.uk/wiki/uploads/HIDExample.zip
// set #option USER_CUSTOM_HID_REPORTS = "UserReportsExample2.bas"
// the HIDExample 'PORTB.0' and 'PORTB.1' checkboxes will control LED1 and LED2
// we'll send a 'Count' message every 1000ms
//-----------------------------------------------------------------------------
// note: because of the long delay calls used, this example only works well
// when using interrupt mode (#option USB_SERVICE = true)
//
#if (DEMO_EXAMPLE = 2)
userInit()      // hdw specific pic init

TXReport.Time = 0

// wait to connect to USB...
low(LED1)
repeat
    USBservice()        // if using interrupts, calls to USBService() will be ignored
until (HID.Attached)
high(LED1)

// main program loop
// use the SF HIDExample.exe tool to change led settings/view messages
while (true)
   // if we have data, set port values, update message
   // and then reset time counter...
    if (HID.DataAvailable()) then
      HID.ReadReport()
      LED1 = RXReport.LED0
      LED2 = RXReport.LED1
      TXReport.Message = "PORT CHANGED"
      HID.WriteReport()
      TXReport.Time = 0
      delayms(100)
   // no data, set waiting message...   
   else
      delayms(1)        
      inc(TXReport.Time)
      if (HID.Attached) then
          if ((TXReport.Time mod 1000) = 0) then    // display msg every 1000ms
            TXReport.Message = "WAITING..."
            HID.WriteReport()
          endif
      endif
   endif
    // if the USB_SERVICE option has been disabled, you MUST poll the USB connection
    // every 1ms or so to keep it alive and handle pending transactions on the bus...
    USBservice()
end while
#endif      // example 2

//
//-----------------------------------------------------------------------------
// example 3 - generic reports
// use the tool MikroE USB HIDTerminal (http://www.mikroe.com/upld/hidterminal.zip)
// comment out #option USER_CUSTOM_HID_REPORTS
//-----------------------------------------------------------------------------
//
#if (DEMO_EXAMPLE = 3)
userInit()      // hdw specific pic init

// wait to connect to USB...
low(LED1)
repeat
    USBservice()        // if using interrupts, calls to USBService() will be ignored
until (HID.Attached)
high(LED1)

// main program loop
// read report messages and send/echo response
while (true)
    if (HID.Attached) then
        high(LED2)
    else
        low(LED2)
    endif

    if (HID.DataAvailable()) then
        high(LED1)                // flash led to indicate we've got a packet
        HID.ReadReport()                // read incoming report
        TXReport.b = RXReport.b         // just echo the received data
        HID.WriteReport()
        low(LED1)
    endif

    // if the USB_SERVICE option has been disabled, you MUST poll the USB connection
    // every 1ms or so to keep it alive and handle pending transactions on the bus...
    USBservice()
end while
#endif      // example 3

end program
