//
//***************************************************************************
// Name    : USBCDC.bas                                                     *
// Author  : David John Barker                                              *
// Notice  : Copyright (c) 2007 Mecanique                                   *
//         : All Rights Reserved                                            *
// Date    : 1/12/2015                                                      *
// Version : 1.4.3                                                          *
//         : adapt for SF protected regions (remove ext/abs allocations)    *
//         : add SendBreak event                                            *
//         : add #option CDC_DOUBLE_BUFFER_TX to allow minimal ram usage    *
// Version : 1.4.2                                                          *
//         : **BUG FIX 1.4-1: correct compile-time check of _CDC_USED_RAM   *
//         :   inadvertantly changed ram buffer locations (check was added  *
//         :   AFTER initial testing), causing issues with limited ram      *
//         :   devices (ie 14K50). add _CDC_TOTAL_USED_RAM instead          *
//         :   rename CDC_VAR_ADDRESS to CDC_BUF_ADDRESS for clarity        *
// Version : 1.4 updated for new stack version                              *
//         : 1.1 ISR now uses USB interrupts                                *
//         : Renamed 'Connected' to 'Attached'                              *
//         : 1.0 Release                                                    *
// Notes   : This is a Virtual COM Port (VCP) implementation using the      *
//         : USB CDC class                                                  *
//***************************************************************************
//
module CDC

// import modules...
#option USB_USE_CDC = true
include "usbmem.bas"
include "usbsfr.bas"
include "usbconfig.bas"
include "usbdefs.bas"
include USB_DESCRIPTOR
include "usbsystem.bas"
include "usbclock.bas"

//
//---------------------------------------------------------------------------
// CDC USB memory requirements
//---------------------------------------------------------------------------
// CDC_USED_RAM = USB_BDT_ENTRY_SIZE * (MAX_EP_NUMBER+1) +  ' = (2*4) * (3+1) = 32
//                sizeof(structure CTRL_TRF_SETUP) +        ' = 8 (USB_EP0_BUFF_SIZE)
//                sizeof(structure CTRL_TRF_DATA) +         ' = 8 (USB_EP0_BUFF_SIZE)
//                CDC_INT_EP_SIZE +                         ' = 8 (aka CDCNotice)
//                CDC_BULK_OUT_EP_SIZE +                    ' = 64 typ (CDC_RX_SIZE)
//                CDC_BULK_IN_EP_SIZE                       ' = 64 typ (CDC_TX_SIZE)
//
// so CDC_USED_RAM = typically 184 bytes
//
#option _CDC_USED_RAM = (USB_BDT_ENTRY_SIZE * (MAX_EP_NUMBER+1)) + 2*USB_EP0_BUFF_SIZE + USB_INT_EP_SIZE + CDC_RX_SIZE + CDC_TX_SIZE
// for inspecting via IDE Code Explorer (you must have 'explore includes' enabled)
public const CDC_USED_RAM = _CDC_USED_RAM

// check to see if USB memory has already been reserved in usbmem.bas
// if not, then resize the region and reserve it here
#if (USBMEM_RESERVE_MEMORY = false)
  #if (SHOW_USB_BUILD_MSGS)
    #if (_CDC_USED_RAM > USB_RAM_SIZE)
      #warning "increasing reserved USB_RAM_SIZE"
    #elseif (_CDC_USED_RAM < USB_RAM_SIZE)
      #warning "decreasing reserved USB_RAM_SIZE"
    #endif
  #endif  // SHOW_USB_BUILD_MSGS
  // resize USB ram to what we require
  #variable USB_RAM_SIZE = _CDC_USED_RAM
  // create protected block to reserve USB ram from SF allocation
  #define _protected_a_start = USB_RAM_BASE
  #define _protected_a_end   = USB_RAM_BASE + (USB_RAM_SIZE-1)
#elseif (_CDC_USED_RAM > USB_RAM_SIZE)  // memory already reserved... make sure it's big enough
  #error "increase USB_RAM_SIZE in usbmem.bas"
#endif      // USBMEM_RESERVE_MEMORY   

// this option controls if Initialize() is called automatically at startup
#option USB_AUTO_INITIALIZE = true

// on control event type
type
    TOnControl = event()

// ISR priority
const
    PRIORITY_LEVEL = USB_SERVICE_PRIORITY

// user buffer sizes (we use endpoint sizes as the defaults)
#option CDC_TXB_SIZE = CDC_TX_SIZE
#option CDC_RXB_SIZE = CDC_RX_SIZE
#option CDC_DOUBLE_BUFFER_TX = true             // true=use two tx buffers, false=one buffer
const
    TX_BUFFER_SIZE = CDC_TXB_SIZE,              // higher value may cause RS232 apps problems, leave!
    RX_BUFFER_SIZE = CDC_RXB_SIZE               // must be >= sizeof(CDCDataRX)

// user application RX and TX buffers
// if option CDC_DOUBLE_BUFFER_TX is true, then two TX buffers are used-
// one is for this module to write to, the other is output by the USB module. 
// by swapping them over for transmission, the module can continue to buffer data 
// while the USB module outputs the last buffer.
// these buffers do not have to be located in USB_RAM since they will be copied
// to/from the actual endpoint buffers
dim
    FRXBuffer(RX_BUFFER_SIZE)  as byte,
    FTXBuffer(TX_BUFFER_SIZE)  as byte

// declare second TX buffer (default)
#if (CDC_DOUBLE_BUFFER_TX)
dim 
    FTXBufferB(TX_BUFFER_SIZE) as byte,
    FTXBufferBPtr as word   // address of second TX buffer
#elseif (isoption(SHOW_USB_BUILD_MSGS) and (SHOW_USB_BUILD_MSGS))
  #warning "CDC using single TX buffer"
#endif

// private variables and aliases
dim
    FTXPos as byte,         // TX buffer current position
    FTXBufferPtr as word,   // address of first TX buffer
    FRXIndexIn as byte,     // RX pointer (received)
    FRXIndexOut as byte,    // RX pointer (read)
    FRXOverrun as boolean   // overrun flag

// public variables...
public dim
    OnControl as TOnControl,               // control event handler
    DTR as USBSystem.CDCControlSignal.DTR, // DTR
    RTS as USBSystem.CDCControlSignal.RTS, // RTS
    Overrun as FRXOverrun,                 // overrun flag
    ReadTerminator as char                 // string read line terminator

// alias for serial break event handle (actually located in USBSystem)
public dim
    SendBreakEvent as USBSystem.CDCSendBreakEvent
    
// NOTE: (from original source)
// It appears that the default windows driver (usbser.sys) does not
// implement a RTS correctly from host to device - this is based on my own
// observations and sources from the internet. However, I have left the above
// RTS mapped to the correct bit, should another driver become available...

public dim
    Attached as USBSystem.Attached

// local clrWDT macro
macro ClrWDT()
     asm
        clrwdt
     end asm
end macro

//
//**************************************************************************
//Name    : ServiceTX (PRIVATE)                                            *
//Purpose : Service the transmission buffer                                *
//**************************************************************************
//
sub ServiceTX()
  #if (CDC_DOUBLE_BUFFER_TX)
    dim temp_ptr as word
    // check if the bulk in EP is available and there's bytes in the user tx buffer
    if (mUSBUSARTIsTxTrfReady() and (FTXPos > 0)) then
        // setup to transfer the user tx buffer to the EP
        // the actual copying will be done by CDCTxService()
        mUSBUSARTTxRam(FTXBufferPtr, FTXPos)
        // swap user buffer ptrs
        temp_ptr = FTXBufferPtr
        FTXBufferPtr = FTXBufferBPtr
        FTXBufferBPtr = temp_ptr
        FTXPos = 0
    endif
  #else     // single TX buffer
    // check if the bulk in EP is available and there's bytes in the user tx buffer
    if (mUSBUSARTIsTxTrfReady() and (FTXPos > 0)) then
        // setup to transfer the user tx buffer to the EP
        // the actual copying will be done by CDCTxService()
        mUSBUSARTTxRam(FTXBufferPtr, FTXPos)
        FTXPos = 0
    endif
  #endif
end sub

//
//**************************************************************************
//Name    : ServiceRX (PRIVATE)                                            *
//Purpose : Service the RX buffer                                          *
//**************************************************************************
//
sub ServiceRX()
    dim BytesRead as byte

    if (mCDCUsartRxIsBusy = 0) then         // there's data in the bulk out EP
        BytesRead = CDCRXBytesAvailable     // get the count of bytes in the EP
        FSR0 = addressof(FRXBuffer)         // set a ptr to the user rx buffer
        inc(FSR0, FRXIndexIn)               // account for bytes already in user buf
        FSR1 = addressof(CDCDataRX)         // set a ptr to the bulk out EP
        repeat                              // copy EP data to our user rx buffer
            POSTINC0 = POSTINC1         
            inc(FRXIndexIn)
            if ((FRXIndexIn > bound(FRXBuffer)) or (FRXIndexIn = 0)) then
                FRXIndexIn = 0
                FSR0 = addressof(FRXBuffer)
            endif
            dec(BytesRead)
            if (not FRXOverrun) then
                FRXOverrun = (FRXIndexIn = FRXIndexOut) and (BytesRead > 0)
            endif
        until (BytesRead = 0)
        CDCRXPrepareForNext()               // pass the EP buffer back to USB SIE
    endif
end sub

//
//**************************************************************************
//Name    : Service                                                        *
//Purpose : Service a USB connection (private)                             *
//**************************************************************************
//
sub _service()
    dim ControlSignal as byte

    ControlSignal = USBSystem.CDCControlSignal._byte
    USBCheckBusStatus()
    USBDriverService()
    CDCTxService()
    ServiceTX()
    ServiceRX()
    ClrWDT()
    if (ControlSignal <> USBSystem.CDCControlSignal._byte) then
        OnControl()
    endif
end sub

//
//**************************************************************************
//Name    : USBService (was private NoISRService)                          *
//Name    : Service                                                        *
//Purpose : Service a USB connection. You need to call this routine        *
//        : periodically from your main program to maintain a connection   *
//        : (every 1 ms or so). You DO NOT need to call this routine if    *
//        : you have enabled the USB interrupt handler, since the routine  *
//        : will be optimized away                                         *
//**************************************************************************
//
public inline sub USBService()
  #if (USB_SERVICE = false)     // only call _service if polled mode
    _service()
  #endif
end sub

//
//**************************************************************************
//Name    : USB_intr_handler (PRIVATE)                                     *
//Purpose : USB isr handler                                                *
//**************************************************************************
// this routine takes ~50us at 48MHz, but since in most cases interrupts are
// generated back to back the effective time is ~100us to 150us.
// they typically occur at 1ms intervals, but that is host dependant
//
#if (USB_SERVICE)
interrupt USB_intr_handler(PRIORITY_LEVEL)
    // check for USB intr request
    if ((USB_IF = 1) and (USB_IE = 1)) then
        USB_IF = 0
        save(0, FSR2, _service)     // save(0) also saves FSR0, FSR1, and PROD
        _service()
        restore
    endif
end interrupt
#endif          // USB_SERVICE

//
//**************************************************************************
//Name    : EnableISR                                                      *
//Purpose : Enable the ISR to service USB connection                       *
//**************************************************************************
//
public inline sub EnableISR()
  #if (USB_SERVICE)
    USB_IE = 1
  #endif
end sub

//
//**************************************************************************
//Name    : DisableISR                                                     *
//Purpose : Disable the ISR that services the USB connection               *
//**************************************************************************
//
public inline sub DisableISR()
  #if (USB_SERVICE)
    USB_IE = 0
  #endif
end sub

//
//**************************************************************************
//Name    : WaitForTX (PRIVATE)                                            *
//Purpose : Wait for next USB TX slot                                      *
//**************************************************************************
//
inline sub WaitForTX()
    while (not mUSBUSARTIsTxTrfReady() and Attached)
        _service()
    end while
end sub

//
//**************************************************************************
//Name    : ClearOverrun                                                   *
//Purpose : Clear RX overrun flag                                          *
//**************************************************************************
//
public sub ClearOverrun()
    DisableISR()
    FRXOverrun = false
    EnableISR()
end sub

//
//**************************************************************************
//Name    : DataAvailable                                                  *
//Purpose : Returns true if RX data is available, false otherwise          *
//**************************************************************************
//
public function DataAvailable() as boolean
    DisableISR()
    result = (FRXIndexIn <> FRXIndexOut)
    EnableISR()
end function

//
//**************************************************************************
//Name    : ReadByte                                                       *
//Purpose : Read a single byte - this is a blocking call                   *
//**************************************************************************
//
public function ReadByte() as byte
    DisableISR()

    while ((FRXIndexIn = FRXIndexOut) and (not FRXOverrun) and Attached)
        _service()
    end while

    result = FRXBuffer(FRXIndexOut)
    inc(FRXIndexOut)
    if (FRXIndexOut > bound(FRXBuffer)) then
        FRXIndexOut = 0
    endif

    _service()
    EnableISR()
end function

//
//**************************************************************************
//Name    : WriteByte                                                      *
//Purpose : Write a byte to the TX buffer - will block if the TX buffer    *
//        : is full until the next transmission slot is available          *
//**************************************************************************
//
public sub WriteByte(pValue as byte)
    DisableISR()

    while ((FTXPos = (TX_BUFFER_SIZE - 1)) and Attached)
        _service()
    end while

  #if (CDC_DOUBLE_BUFFER_TX = false)
    // if we've already started sending data from the TXbuffer then
    // wait for it to finish
    WaitForTX()
  #endif
    
    FSR0 = FTXBufferPtr
    inc(FSR0, FTXPos)
    INDF0 = pValue
    inc(FTXPos)

    _service()
    EnableISR()
end sub

//
//**************************************************************************
//Name    : ReadArray                                                      *
//Purpose : Read an array of bytes - pCount is an optional number of bytes *
//        : to read. If pCount is not given, the size of the array is used *
//**************************************************************************
//
public sub ReadArray(byref pArray() as byte, pCount as word = 0)
    if (pCount = 0) then
        pCount = bound(pArray)
        inc(pCount)
    endif

    DisableISR()
    FSR2 = addressof(pArray)
    while ((pCount > 0) and Attached)
        if ((FRXIndexIn = FRXIndexOut) and (not FRXOverrun)) then
            _service()
        else
            POSTINC2 = FRXBuffer(FRXIndexOut)
            inc(FRXIndexOut)
            if (FRXIndexOut > bound(FRXBuffer)) then
                FRXIndexOut = 0
            endif
            dec(pCount)
        endif
    end while
    EnableISR()
end sub

//
//**************************************************************************
//Name    : WriteArray                                                     *
//Purpose : Write an array of bytes - pCount is an optional number of      *
//        : to write. If pCount is not given, the array size is used       *
//**************************************************************************
//
public sub WriteArray(byref pArray() as byte, pCount as word = 0)
    if (pCount = 0) then
        pCount = bound(pArray)
        inc(pCount)
    endif

    DisableISR()
    WaitForTX()
    mUSBUSARTTxRam(addressof(pArray), pCount)
    EnableISR()
end sub

//
//**************************************************************************
//Name    : ReadItem (OVERLOAD)                                            *
//Purpose : Read a byte from the USB CDC connection                        *
//**************************************************************************
//
sub ReadItem(byref pValue as byte)
    pValue = ReadByte()
end sub

//
//**************************************************************************
//Name    : ReadItem (OVERLOAD)                                            *
//Purpose : Read a string from the USB CDC connection. Use ReadTerminator  *
//        : to specify the input string terminator character.              *
//**************************************************************************
//
sub ReadItem(byref pText as string)
    dim b as byte

    FSR2 = addressof(pText)
    b = ReadByte()
    while (b <> byte(ReadTerminator))
        POSTINC2 = b
        b = ReadByte()
    end while
    POSTINC2 = null
end sub

//
//**************************************************************************
//Name    : ReadItem (OVERLOAD)                                            *
//Purpose : Read an array from the USB CDC connection                      *
//**************************************************************************
//
sub ReadItem(byref pArray() as byte)
    ReadArray(pArray)
end sub

//
//**************************************************************************
//Name    : Read (COMPOUND)                                                *
//Purpose : Read an item from the USB CDC connection                       *
//**************************************************************************
//
public compound sub Read(ReadItem)

//
//**************************************************************************
//Name    : WriteItem (OVERLOAD)                                           *
//Purpose : Write a byte value to the USB CDC connection                   *
//**************************************************************************
//
sub WriteItem(pValue as WREG)
    WriteByte(pValue)
end sub

//
//**************************************************************************
//Name    : WriteItem (OVERLOAD)                                           *
//Purpose : Write a string value to the USB CDC connection                 *
//**************************************************************************
//
sub WriteItem(pText as string)
    FSR2 = addressof(pText)
    while (INDF2 <> 0)
        WriteByte(POSTINC2)
    end while
end sub

//
//**************************************************************************
//Name    : WriteItem (OVERLOAD)                                           *
//Purpose : Write an array value to the USB CDC connection                 *
//**************************************************************************
//
sub WriteItem(byref pArray() as byte)
    WriteArray(pArray)
end sub

//
//**************************************************************************
//Name    : Write (COMPOUND)                                               *
//Purpose : Write an item to the USB CDC connection                        *
//**************************************************************************
//
public compound sub Write(WriteItem)

//
//**************************************************************************
//Name    : WaitFor                                                        *
//Purpose : Wait for a byte value to be received                           *
//**************************************************************************
//
public function WaitFor(pValue as byte) as boolean
    result = ReadByte() = pValue
end function

//
//**************************************************************************
//Name    : DataAvailableTimeout                                           *
//Purpose : Checks to see if a byte value has been received, with          *
//        : timeout in milliseconds                                        *
//**************************************************************************
//
public function DataAvailableTimeout(pTimeout as word) as boolean
    while ((pTimeout > 0) and (not DataAvailable()))
        delayms(1)
        USBService()
        dec(pTimeout)
    end while
    result = DataAvailable()
end function

//
//**************************************************************************
//Name    : Initialize                                                     *
//Purpose : Initialize the USB module                                      *
//**************************************************************************
//
// initialize module variables
inline sub InitializeCDC()
    FTXPos = 0
    FTXBufferPtr = addressof(FTXBuffer)
  #if (CDC_DOUBLE_BUFFER_TX)
    FTXBufferBPtr = addressof(FTXBufferB)
  #endif
    FRXIndexIn = 0
    FRXIndexOut = 0
    FRXOverrun = false
    ReadTerminator = null
    OnControl = 0
end sub

public sub Initialize()
    // setup usb clock (see USBClock.bas)
    USBClockInit()

    mInitializeUSBDriver()

  #if (USB_SERVICE)
    disable(USB_intr_handler)
    #if (USB_SERVICE_PRIORITY = ipLow)
      USB_IP = 0
    #else
      USB_IP = 1
    #endif
    USB_IF = 0      // clear USBIF
    USB_IE = 1      // enable USB interrupt

    _service()
    UIE = UIE_ENABLE_ALL    // unmask all USB interrupts
    UEIE = UEIE_ENABLE_ALL  // unmask all USB error interrupts
    enable(USB_intr_handler)
  #endif    // USB_SERVICE
end sub

//
//**************************************************************************
// initialize the module
//**************************************************************************
//
InitializeCDC()
#if (USB_AUTO_INITIALIZE)
  Initialize()
#elseif (SHOW_USB_BUILD_MSGS)
  #warning "USB_AUTO_INITIALIZE false... driver NOT initialized at startup"
#endif

end module
