//
//***************************************************************************
// Name    : USBDefs.bas                                                    *
// Author  : David John Barker                                              *
// Notice  : Copyright (c) 2007 Mecanique                                   *
//         : All Rights Reserved                                            *
// Date    : 10/01/2007                                                     *
// Version : 1.4 minor editing changes                                      *
//         : 1.2 Microchip firmware bug fix - search for 'bug fix' in this  *
//         : file                                                           *
//         : 1.1 Corrected PID declaration in BD_STAT                       *
//         : 1.0 Release                                                    *
// Notes   : Constants and structures used by USBSystem.bas                 *
//***************************************************************************
//
module USBDefs

// import module...
include "usbconfig.bas"

// bring #options into module...
#define USB_BDT_ENTRY_SIZE = 8  // each BD uses 4 bytes, so x2 assuming both IN and OUT

// EP0_BUFF_SIZE defines the buffer size for endpoint 0. It can have a valid value of 8,
// 16, 32 or 64 bytes. This definition is used globally in the project for many things.
// It is used during project build to allocate appropriate buffer size for endpoint 0.
// It is used in the USB descriptor to notify the USB host of the size of the endpoint 0
// buffer. It is also used during control transfer operation.
// When defining this variable, note that a low-speed USB device can only use an 8-byte
// buffer, while a full-speed USB device can use an 8, 16, 32 or 64-byte buffer.
#define USB_EP0_BUFF_SIZE = 8

// size of a USB INTERRUPT endpoint
#define USB_INT_EP_SIZE = 8

public const
   EP0_BUFF_SIZE       = USB_EP0_BUFF_SIZE,
   HID_REPORT_IN_SIZE  = HID_REPORT_BYTES_IN,
   HID_REPORT_OUT_SIZE = HID_REPORT_BYTES_OUT,
   USBConnectMS        = USB_CONNECT_MS

// usbcfg.h (CDC)
public const
   CDC_COMM_INTF_ID =       $00,
   CDC_DATA_INTF_ID =       $01,
   CDC_INT_EP_SIZE =        USB_INT_EP_SIZE,    // CDCNotice
   CDC_BULK_OUT_EP_SIZE =   CDC_RX_SIZE,        // CDCDataRX
   CDC_BULK_IN_EP_SIZE =    CDC_TX_SIZE         // CDCDataTX

// UCFG Initialization Parameters
public const                                                                    'usbdrv.h
   _PPBM0 =     $00,           // Pingpong Buffer Mode 0
   _PPBM1 =     $01,           // Pingpong Buffer Mode 1
   _PPBM2 =     $02,           // Pingpong Buffer Mode 2
   _LS =        $00,           // Use Low-Speed USB Mode
   _FS =        $04,           // Use Full-Speed USB Mode
   _TRINT =     $00,           // Use internal transceiver
   _TREXT =     $08,           // Use external transceiver
   _PUEN =      $10,           // Use internal pull-up resistor
   _OEMON =     $40,           // Use SIE output indicator
   _UTEYE =     $80            // Use Eye-Pattern test

// UEPn Initialization Parameters
public const                                                                    'usbdrv.h
   EP_CTRL =    $06,           // Cfg Control pipe for this ep
   EP_OUT =     $0C,           // Cfg OUT only pipe for this ep
   EP_IN =      $0A,           // Cfg IN only pipe for this ep
   EP_OUT_IN =  $0E,           // Cfg both OUT & IN pipes for this ep
   HSHK_EN =    $10            // Enable handshake packet
                               // Handshake should be disable for isoch
//
// USB - PICmicro Endpoint Definitions
// This is used when checking the value read from USTAT
// NOTE: These definitions are not used in the descriptors.
// EP addresses used in the descriptors have different format and
// are defined in usbdefs_std_dsc.h
//
public const                                                                    'usbdrv.h
   OUT =         0,
   IN =          1,
   PIC_EP_NUM_MASK = %01111000,
   PIC_EP_DIR_MASK = %00000100,
   EP00_OUT =   ($00<<3) or (OUT<<2),
   EP00_IN =    ($00<<3) or (IN<<2),
   EP01_OUT =   ($01<<3) or (OUT<<2),
   EP01_IN =    ($01<<3) or (IN<<2),
   EP02_OUT =   ($02<<3) or (OUT<<2),
   EP02_IN =    ($02<<3) or (IN<<2),
   EP03_OUT =   ($03<<3) or (OUT<<2),
   EP03_IN =    ($03<<3) or (IN<<2),
   EP04_OUT =   ($04<<3) or (OUT<<2),
   EP04_IN =    ($04<<3) or (IN<<2),
   EP05_OUT =   ($05<<3) or (OUT<<2),
   EP05_IN =    ($05<<3) or (IN<<2),
   EP06_OUT =   ($06<<3) or (OUT<<2),
   EP06_IN =    ($06<<3) or (IN<<2),
   EP07_OUT =   ($07<<3) or (OUT<<2),
   EP07_IN =    ($07<<3) or (IN<<2),
   EP08_OUT =   ($08<<3) or (OUT<<2),
   EP08_IN =    ($08<<3) or (IN<<2),
   EP09_OUT =   ($09<<3) or (OUT<<2),
   EP09_IN =    ($09<<3) or (IN<<2),
   EP10_OUT =   ($0A<<3) or (OUT<<2),
   EP10_IN =    ($0A<<3) or (IN<<2),
   EP11_OUT =   ($0B<<3) or (OUT<<2),
   EP11_IN =    ($0B<<3) or (IN<<2),
   EP12_OUT =   ($0C<<3) or (OUT<<2),
   EP12_IN =    ($0C<<3) or (IN<<2),
   EP13_OUT =   ($0D<<3) or (OUT<<2),
   EP13_IN =    ($0D<<3) or (IN<<2),
   EP14_OUT =   ($0E<<3) or (OUT<<2),
   EP14_IN =    ($0E<<3) or (IN<<2),
   EP15_OUT =   ($0F<<3) or (OUT<<2),
   EP15_IN =    ($0F<<3) or (IN<<2)

public const                                                                    'usbdefs_std_dsc.h
   DSC_DEV =  $01,
   DSC_CFG =  $02,
   DSC_STR =  $03,
   DSC_INTF = $04,
   DSC_EP =   $05

//
// USB Endpoint Definitions
// This is used in the descriptors.
// NOTE: Do not use these values for checking against USTAT.
// To check against USTAT, use values defined in usbdrv.h
//
public const                                                                    'usbdefs_std_dsc.h
   _EP01_OUT =  $01,
   _EP01_IN  =  $81,
   _EP02_OUT =  $02,
   _EP02_IN  =  $82,
   _EP03_OUT =  $03,
   _EP03_IN  =  $83,
   _EP04_OUT =  $04,
   _EP04_IN  =  $84,
   _EP05_OUT =  $05,
   _EP05_IN  =  $85,
   _EP06_OUT =  $06,
   _EP06_IN  =  $86,
   _EP07_OUT =  $07,
   _EP07_IN  =  $87,
   _EP08_OUT =  $08,
   _EP08_IN  =  $88,
   _EP09_OUT =  $09,
   _EP09_IN  =  $89,
   _EP10_OUT =  $0A,
   _EP10_IN  =  $8A,
   _EP11_OUT =  $0B,
   _EP11_IN  =  $8B,
   _EP12_OUT =  $0C,
   _EP12_IN  =  $8C,
   _EP13_OUT =  $0D,
   _EP13_IN  =  $8D,
   _EP14_OUT =  $0E,
   _EP14_IN  =  $8E,
   _EP15_OUT =  $0F,
   _EP15_IN  =  $8F

// configuration attributes
public const                                                                    'usbdefs_std_dsc.h
   _DEFAULT =   $01 << 7,        //Default Value (Bit 7 is set)
   _SELF =      $01 << 6,        //Self-powered (Supports if set)
   _RWU =       $01 << 5         //Remote Wakeup (Supports if set)

// endpoint transfer type
public const                                                                    'usbdefs_std_dsc.h
   _CTRL =      $00,            //Control Transfer
   _ISO =       $01,            //Isochronous Transfer
   _BULK =      $02,            //Bulk Transfer
   _INT =       $03             //Interrupt Transfer

// isochronous endpoint synchronization type                                    'usbdefs_std_dsc.h
public const
   _NS =        $00 << 2,       //No Synchronization
   _AS =        $01 << 2,       //Asynchronous
   _AD =        $02 << 2,       //Adaptive
   _SY =        $03 << 2        //Synchronous

// isochronous endpoint usage type
public const                                                                    'usbdefs_std_dsc.h
   _DE =        $00 << 4,       //Data endpoint
   _FE =        $01 << 4,       //Feedback endpoint
   _IE =        $02 << 4        //Implicit feedback Data endpoint

// MAX_NUM_INTFC defines the size of the array which keeps track of the active
// alternate setting for each interface, which can be switched during operation.
// Valid values are integers [0, 1, 2,...]. If a device has multiple configurations,
// the number of interfaces from the configuration with the highest number of interfaces
// should be used. For example, a device with two configurations has three interfaces
// in the first configuration and two interfaces in the second. In this case, MAX_NUM_INTFC
// should be three
public const
   MAX_NUM_INTFC = 1                                                            'usbcfg.h

// UCFG_VAL defines the initial value for the UCFG Special Function Register
// Parameter definitions are defined above
public const                                                                    'usbcfg.h
   MODE_PP =   _PPBM0,
   UCFG_VAL =  _PUEN or _TRINT or _FS or MODE_PP

//
// MUID = Microchip USB Class ID
// Used to identify which of the USB classes owns the current
// session of control transfer over EP0
//
public const                                                                    'usbcfg.h
   MUID_NULL = 0,
   MUID_USB9 = 1,
   MUID_HID  = 2,
   MUID_CDC  = 3,
   MUID_MSD  = 4

// endpoints allocation - see usbmmap.h for an explanation of how the endpoint
// allocation works
public const                                                                    'usbcfg.h
   HID_INTF_ID =          0,
   HID_INT_OUT_EP_SIZE =  HID_REPORT_IN_SIZE,
   HID_INT_IN_EP_SIZE =   HID_REPORT_OUT_SIZE,
   HID_NUM_OF_DSC =       1,
   HID_RPT01_SIZE =       47

// the number of bytes in each report,
// calculated from ReportSize and ReportCount in the report descriptor
public const                                                                    'usbcfg.h
   HID_INPUT_REPORT_BYTES =   HID_REPORT_IN_SIZE,
   HID_OUTPUT_REPORT_BYTES =  HID_REPORT_OUT_SIZE,
   HID_FEATURE_REPORT_BYTES = 2

// buffer descriptor status register initialization parameters
public const                                                                    'usbmap.h
   _BSTALL =    $04,                //Buffer Stall enable
   _DTSEN =     $08,                //Data Toggle Synch enable
   _INCDIS =    $10,                //Address increment disable
   _KEN =       $20,                //SIE keeps buff descriptors enable
   _DAT0 =      $00,                //DATA0 packet expected next
   _DAT1 =      $40,                //DATA1 packet expected next
   _DTSMASK =   $40,                //DTS Mask
   _USIE =      $80,                //SIE owns buffer
   _UCPU =      $00                 //CPU owns buffer

// USB device states - To be used with usb_device_state
public const                                                                    'usbmap.h
   DETACHED_STATE =    0,
   ATTACHED_STATE =    1,
   POWERED_STATE =     2,
   DEFAULT_STATE =     3,
   ADR_PENDING_STATE = 4,
   ADDRESS_STATE =     5,
   CONFIGURED_STATE =  6

// memory types for control transfer - used in USB_DEVICE_STATUS
public const                                                                    'usbmap.h
   _RAM = 0,
   _ROM = 1

// Standard Request Codes
// USB 2.0 Spec Ref Table 9-4
public const                                                                    'usb9.h
   GET_STATUS =  0,
   CLR_FEATURE = 1,
   SET_FEATURE = 3,
   SET_ADR =     5,
   GET_DSC =     6,
   SET_DSC =     7,
   GET_CFG =     8,
   SET_CFG =     9,
   GET_INTF =    10,
   SET_INTF =    11,
   SYNCH_FRAME = 12

// standard feature selectors
public const                                                                    'usb9.h
   DEVICE_REMOTE_WAKEUP = $01,
   ENDPOINT_HALT =        $00

// control transfer states
public const                                                                    'usbctrltr.h
   WAIT_SETUP =  0,
   CTRL_TRF_TX = 1,
   CTRL_TRF_RX = 2

//
//********************************************************************
//Bug Fix: May 14, 2007 (#F7)
//********************************************************************
//For a control transfer read, if the host tries to read more data
//than what it has requested, the peripheral device should stall the
//extra IN transactions and the status stage. Typically, a host does
//not try to read more data than what it has requested. The original
//firmware did not handle this situation. Instead of stalling extra
//IN transactions, the device kept sending out zero length packets.
//
//The new definitions introduced is used to keep track if a short IN
//packet has been sent or not. From this the state machine can
//decide if it should stall future extra IN transactions or not.
//********************************************************************
//
// Short Packet States - Used by Control Transfer Read  - CTRL_TRF_TX
public const
   SHORT_PKT_NOT_USED  = 0,
   SHORT_PKT_PENDING   = 1,
   SHORT_PKT_SENT      = 2

// USB PID: Token Types - See chapter 8 in the USB specification
public const                                                                    'usbctrltr.h
   SETUP_TOKEN = %00001101,
   OUT_TOKEN =   %00000001,
   IN_TOKEN =    %00001001

// bmRequestType definitions
public const                                                                    'usbctrltr.h
   HOST_TO_DEV = 0,
   DEV_TO_HOST = 1

public const                                                                    'usbctrltr.h
   STANDARD =    $00,
   CLASS =       $01,
   VENDOR =      $02

public const                                                                    'usbctrltr.h
   RCPT_DEV =    0,
   RCPT_INTF =   1,
   RCPT_EP =     2,
   RCPT_OTH =    3

// class-specific requests
public const                                                                    'hid.h
   GET_REPORT =     $01,
   GET_IDLE =       $02,
   GET_PROTOCOL =   $03,
   SET_REPORT =     $09,
   SET_IDLE =       $0A,
   SET_PROTOCOL =   $0B

// class descriptor types
public const                                                                    'hid.h
   DSC_HID =        $21,
   DSC_RPT =        $22,
   DSC_PHY =        $23

// protocol selection
public const                                                                    'hid.h
   BOOT_PROTOCOL =  $00,
   RPT_PROTOCOL =   $01

// HID interface class code
public const                                                                    'hid.h
   HID_INTF =       $03

// HID interface class subclass codes
public const                                                                    'hid.h
   BOOT_INTF_SUBCLASS = $01

// HID interface class protocol codes
public const                                                                    'hid.h
   HID_PROTOCOL_NONE =      $00,
   HID_PROTOCOL_KEYBOAD =   $01,
   HID_PROTOCOL_MOUSE =     $02

public const                                                                    'cdc.h
   SEND_ENCAPSULATED_COMMAND =  $00,
   GET_ENCAPSULATED_RESPONSE =  $01,
   SET_COMM_FEATURE =           $02,
   GET_COMM_FEATURE =           $03,
   CLEAR_COMM_FEATURE =         $04,
   SET_LINE_CODING =            $20,
   GET_LINE_CODING =            $21,
   SET_CONTROL_LINE_STATE =     $22,
   SEND_BREAK =                 $23

// Note: Notifications are polled over Communication Interface (Interrupt Endpoint)
public const                                                                    'cdc.h
   NETWORK_CONNECTION =         $00,
   RESPONSE_AVAILABLE =         $01,
   SERIAL_STATE =               $20

// Device Class Code
public const                                                                    'cdc.h
   CDC_DEVICE =                 $02

// Communication Interface Class Code
public const                                                                    'cdc.h
   COMM_INTF =                  $02

// Communication Interface Class SubClass Codes
public const                                                                    'cdc.h
   ABSTRACT_CONTROL_MODEL =     $02

// Communication Interface Class Control Protocol Codes
public const                                                                    'cdc.h
   V25TER =                     $01    // Common AT commands ("Hayes(TM)")

// Data Interface Class Codes
public const                                                                    'cdc.h
   DATA_INTF =                  $0A

// Data Interface Class Protocol Codes                                          'cdc.h
public const
   NO_PROTOCOL =                $00    // No class specific protocol required

// Communication Feature Selector Codes
public const                                                                    'cdc.h
   ABSTRACT_STATE =             $01,
   COUNTRY_SETTING =            $02

// Functional Descriptors
// Type Values for the bDscType Field
public const                                                                    'cdc.h
   CS_INTERFACE =               $24,
   CS_ENDPOINT =                $25

// bDscSubType in Functional Descriptors
public const                                                                    'cdc.h
   DSC_FN_HEADER =              $00,
   DSC_FN_CALL_MGT =            $01,
   DSC_FN_ACM =                 $02,    // ACM - Abstract Control Management
   DSC_FN_DLM =                 $03,    // DLM - Direct Line Managment
   DSC_FN_TELEPHONE_RINGER =    $04,
   DSC_FN_RPT_CAPABILITIES =    $05,
   DSC_FN_UNION =               $06,
   DSC_FN_COUNTRY_SELECTION =   $07,
   DSC_FN_TEL_OP_MODES =        $08,
   DSC_FN_USB_TERMINAL =        $09
   // more.... see Table 25 in USB CDC Specification 1.1

// CDC Bulk IN transfer states
public const
   CDC_TX_READY =               0,                                              'cdc.h
   CDC_TX_BUSY =                1,
   CDC_TX_BUSY_ZLP =            2,       // ZLP: Zero Length Packet
   CDC_TX_COMPLETING =          3

// pointer type
public structure POINTER                                                        'typedefs.h
   _word as word
   bLow     as _word.byte0
   bHigh    as _word.byte1
   bRam     as _word
   wRam     as _word
   bRom     as _word
   wRom     as _word
end structure

// device status structure
public structure USB_DEVICE_STATUS                                              'usbmap.h
   _byte as byte
   RemoteWakeup as _byte.bits(0)    // [0]Disabled [1]Enabled: See usbdrv.c,usb9.c
   ctrl_trf_mem as _byte.bits(1)    // [0]RAM      [1]ROM
end structure

// endpoint status structure
public structure BD_STAT                                                        'usbmap.h
   _byte as byte
   BC8      as _byte.bits(0)
   BC9      as _byte.bits(1)
   BSTALL   as _byte.bits(2)    // Buffer Stall Enable
   DTSEN    as _byte.bits(3)    // Data Toggle Synch Enable
   INCDIS   as _byte.bits(4)    // Address Increment Disable
   KEN      as _byte.bits(5)    // BD Keep Enable
   DTS      as _byte.bits(6)    // Data Toggle Synch Value
   UOWN     as _byte.bits(7)    // USB Ownership

   PID0     as _byte.bits(2)
   PID1     as _byte.bits(3)
   PID2     as _byte.bits(4)
   PID3     as _byte.bits(5)

   PID      as _byte            // Packet Identifier
end structure

// endpoint structure
public structure BDT                                                            'usbmap.h
   Stat as BD_STAT
   Cnt  as byte
   ADR  as word
   ADRL as ADR.byte0        // Buffer Address Low
   ADRH as ADR.byte1        // Buffer Address High
end structure

// control transfer setup structure
public structure CTRL_TRF_SETUP                                                 'usbdefs_epo_buff.h
   _byte(EP0_BUFF_SIZE) as byte

   bmRequestType    as _byte(0)
   bRequest         as _byte(1)
   wValue           as _byte(2).AsWord
   wIndex           as _byte(4).AsWord
   wLength          as _byte(6).AsWord

   W_Value          as _byte(2).AsWord
   W_Index          as _byte(4).AsWord
   W_Length         as _byte(6).AsWord

   Recipient_f5     as _byte(0)         // Device, Interface, Endpoint, Other (0..4)
   RequestType_f2   as _byte(0)         // Standard, Class, Vendor, Reserved  (5..6)
   DataDir          as _byte(0).bits(7) // Host-to-device, Device-to-host
   bFeature         as _byte(2)         // DEVICE_REMOTE_WAKEUP, ENDPOINT_HALT

   bDscIndex        as _byte(2)         // For Configuration and String DSC Only
   bDscType         as _byte(3)         // Device, Configuration, String
   wLangID          as _byte(4).AsWord  // Language ID

   bDevADR          as _byte(2)         // Device Address 0-127
   bDevADRH         as _byte(3)         // Must equal zero

   bCfgValue        as _byte(2)         // Configuration Value 0-255
   bCfgRSD          as _byte(3)         // Must equal zero (Reserved)

   bAltID           as _byte(2)         // Alternate Setting Value 0-255
   bAltID_H         as _byte(3)         // Must equal zero
   bIntfID          as _byte(4)         // Interface Number Value 0-255
   bIntfID_H        as _byte(5)         // Must equal zero

   bEPID            as _byte(4)         // Endpoint ID (Number & Direction)
   bEPID_H          as _byte(5)         // Must equal zero

   EPNum_f4         as _byte(4)         // Endpoint Number 0-15 (0..3)
   EPDir            as _byte(4).bits(7) // Endpoint Direction: 0-OUT, 1-IN
end structure

// control transfer data structure
public structure CTRL_TRF_DATA                                                  'usbdefs_epo_buff.h
   _byte(EP0_BUFF_SIZE) as byte

   _byte0 as _byte(0)
   _byte1 as _byte(1)
   _byte2 as _byte(2)
   _byte3 as _byte(3)
   _byte4 as _byte(4)
   _byte5 as _byte(5)
   _byte6 as _byte(6)
   _byte7 as _byte(7)

   _word0 as _byte(0).AsWord
   _word1 as _byte(2).AsWord
   _word2 as _byte(4).AsWord
   _word3 as _byte(6).AsWord
end structure

// device descriptor structure
public structure USB_DEV_DSC                                                    'usbdefs_std_dsc.h
    bLength         as byte
    bDscType        as byte
    bcdUSB          as word
    bDevCls         as byte
    bDevSubCls      as byte
    bDevProtocol    as byte
    bMaxPktSize0    as byte
    idVendor        as word
    idProduct       as word
    bcdDevice       as word
    iMFR            as byte
    iProduct        as byte
    iSerialNum      as byte
    bNumCfg         as byte
end structure
public const USB_DEV_DSC_SIZE = sizeof(USB_DEV_DSC)

// configuration descriptor structure
public structure USB_CFG_DSC                                                    'usbdefs_std_dsc.h
    bLength         as byte
    bDscType        as byte
    wTotalLength    as word
    bNumIntf        as byte
    bCfgValue       as byte
    iCfg            as byte
    bmAttributes    as byte
    bMaxPower       as byte
end structure

// interface descriptor structure
public structure USB_INTF_DSC                                                   'usbdefs_std_dsc.h
    bLength         as byte
    bDscType        as byte
    bIntfNum        as byte
    bAltSetting     as byte
    bNumEPs         as byte
    bIntfCls        as byte
    bIntfSubCls     as byte
    bIntfProtocol   as byte
    iIntf           as byte
end structure

// enpoint descriptor structure
public structure USB_EP_DSC                                                     'usbdefs_std_dsc.h
    bLength         as byte
    bDscType        as byte
    bEPAdr          as byte
    bmAttributes    as byte
    wMaxPktSize     as word
    bInterval       as byte
end structure

// HID descriptor header structure
public structure USB_HID_DSC_HEADER                                             'hid.h
   bDscType     as byte
   wDscLength   as word
end structure

// HID descriptor structure
public structure USB_HID_DSC                                                    'hid.h
   bLength      as byte
   bDscType     as byte
   bcdHID       as word
   bCountryCode as byte
   bNumDsc      as byte
   hid_dsc_header(HID_NUM_OF_DSC) as USB_HID_DSC_HEADER
end structure

// Line Coding Structure
public const                                                                    'cdc.h
   LINE_CODING_LENGTH = $07

public structure CDC_LINE_CODING                                                'cdc.h
   _byte(LINE_CODING_LENGTH) as byte
   dwDTERate    as _byte(0).AsLongWord          // Complex data structure
   bCharFormat  as _byte(4)
   bParityType  as _byte(5)
   bDataBits    as _byte(6)
end structure

public structure CDC_CONTROL_SIGNAL_BITMAP                                      'cdc.h
   _byte as byte
   DTR      as _byte.bits(0)
   RTS      as _byte.bits(1)
end structure

// Functional Descriptor Structure - See CDC Specification 1.1 for details
public structure USB_CDC_HEADER_FN_DSC                                          'cdc.h
   bFNLength    as byte
   bDscType     as byte
   bDscSubType  as byte
   bcdCDC       as word
end structure

// Abstract Control Management Functional Descriptor
public structure USB_CDC_ACM_FN_DSC                                             'cdc.h
   bFNLength        as byte
   bDscType         as byte
   bDscSubType      as byte
   bmCapabilities   as byte
end structure

// Union Functional Descriptor
public structure USB_CDC_UNION_FN_DSC                                           'cdc.h
   bFNLength    as byte
   bDscType     as byte
   bDscSubType  as byte
   bMasterIntf  as byte
   bSaveIntf0   as byte
end structure

// Call Management Functional Descriptor
public structure USB_CDC_CALL_MGT_FN_DSC                                        'cdc.h
   bFNLength        as byte
   bDscType         as byte
   bDscSubType      as byte
   bmCapabilities   as byte
   bDataInterface   as byte
end structure

end module
