//
//***************************************************************************
// Name    : USBHID.bas                                                     *
// Author  : David John Barker                                              *
// Notice  : Copyright (c) 2007 Mecanique                                   *
//         : All Rights Reserved                                            *
// Date    : 01/06/2015                                                     *
// Version : 1.4.3                                                          *
//         : adapt for SF protected regions (remove ext/abs allocations)    *
//         : add custom reports for ReadReport()/WriteReport()              *
//         : 1.1 ISR now uses USB interrupts                                *
//         : Renamed 'Connected' to 'Attached'                              *
//         : 1.0 Release                                                    *
// Notes   : This HID module provides a simple way for you to transfer      *
//         : blocks of data to and from your PC using the USB HID class.    *
//***************************************************************************
//
module HID

// import USB modules...
#option USB_USE_HID = true
include "usbmem.bas"
include "usbsfr.bas"
include "usbconfig.bas"
include "usbdefs.bas"
include USB_DESCRIPTOR
include "usbsystem.bas"
include "usbclock.bas"

//
//---------------------------------------------------------------------------
// HID USB memory requirements
//---------------------------------------------------------------------------
// HID_USED_RAM = USB_BDT_ENTRY_SIZE * (MAX_EP_NUMBER+1) +  ' = (2*4) * (1+1) = 16
//                sizeof(structure CTRL_TRF_SETUP) +        ' = 8 (USB_EP0_BUFF_SIZE)
//                sizeof(structure CTRL_TRF_DATA) +         ' = 8 (USB_EP0_BUFF_SIZE)
//                sizeof(hid_report_in) +                   ' = 64 (HID_REPORT_BYTES_IN)
//                sizeof(hid_report_out) +                  ' = 64 (HID_REPORT_BYTES_OUT)
//                sizeof(hid_report_feature)                ' = 8 (USB_EP0_BUFF_SIZE)
//
// so HID_USED_RAM = typically 168 bytes
//
#option _HID_USED_RAM = (USB_BDT_ENTRY_SIZE * (MAX_EP_NUMBER+1)) + 2*USB_EP0_BUFF_SIZE + HID_REPORT_BYTES_IN + HID_REPORT_BYTES_OUT + USB_EP0_BUFF_SIZE
// for inspecting via IDE Code Explorer (you must have 'explore includes' enabled)
public const HID_USED_RAM = _HID_USED_RAM

// check to see if USB memory has already been reserved in usbmem.bas
// if not, then resize the region and reserve it here
#if (USBMEM_RESERVE_MEMORY = false)
  #if (SHOW_USB_BUILD_MSGS)
    #if (_HID_USED_RAM > USB_RAM_SIZE)
      #warning "increasing reserved USB_RAM_SIZE"
    #elseif (_HID_USED_RAM < USB_RAM_SIZE)
      #warning "decreasing reserved USB_RAM_SIZE"
    #endif
  #endif  // SHOW_USB_BUILD_MSGS
  // resize USB ram to what we require
  #variable USB_RAM_SIZE = _HID_USED_RAM
  // create protected block to reserve USB ram from SF allocation
  #define _protected_a_start = USB_RAM_BASE
  #define _protected_a_end   = USB_RAM_BASE + (USB_RAM_SIZE-1)
#elseif (_HID_USED_RAM > USB_RAM_SIZE)  // memory already reserved... make sure it's big enough
  #error "increase USB_RAM_SIZE in usbmem.bas"
#endif      // USBMEM_RESERVE_MEMORY   
      
// this option controls if Initialize() is called automatically at startup
#option USB_AUTO_INITIALIZE = true

// ISR priority
const
    PRIORITY_LEVEL = USB_SERVICE_PRIORITY

public dim
    Attached as USBSystem.Attached

// local clrWDT macro
macro ClrWDT()
     asm
        clrwdt
     end asm
end macro

//
//**************************************************************************
//Name    : Service                                                        *
//Purpose : Service a USB connection (private)                             *
//**************************************************************************
//
sub _service()
    USBCheckBusStatus()
    USBDriverService()
    ClrWDT()
end sub

//
//**************************************************************************
//Name    : USBService (was private NoISRService)                          *
//Name    : Service                                                        *
//Purpose : Service a USB connection. You need to call this routine        *
//        : periodically from your main program to maintain a connection   *
//        : (every 1 ms or so). You DO NOT need to call this routine if    *
//        : you have enabled the USB interrupt handler, since the routine  *
//        : will be optimized away                                         *
//**************************************************************************
//
public inline sub USBService()
  #if (USB_SERVICE = false)     // only call _service if polled mode
    _service()
  #endif
end sub

//
//**************************************************************************
//Name    : USB_intr_handler (PRIVATE)                                     *
//Purpose : USB isr handler                                                *
//**************************************************************************
// this routine takes ~20us at 48MHz, but since in most cases interrupts are 
// generated back to back the effective time is ~40us. 
// they occur at 1ms intervals, but that is host dependant
//
#if (USB_SERVICE)
interrupt USB_intr_handler(PRIORITY_LEVEL)
    // check for USB intr request
    if ((USB_IF = 1) and (USB_IE = 1)) then
        USB_IF = 0
        save(FSR0, FSR1, PRODL, PRODH, _service)
        _service()
        restore
    endif
end interrupt
#endif          // USB_SERVICE

//
//**************************************************************************
//Name    : EnableISR                                                      *
//Purpose : Enable the ISR to service USB connection                       *
//**************************************************************************
//
public inline sub EnableISR()
  #if (USB_SERVICE)
    USB_IE = 1
  #endif
end sub

//
//**************************************************************************
//Name    : DisableISR                                                     *
//Purpose : Disable the ISR that services the USB connection               *
//**************************************************************************
//
public inline sub DisableISR()
  #if (USB_SERVICE)
    USB_IE = 0
  #endif
end sub

//
//**************************************************************************
//Name    : WaitForTX (was PRIVATE)                                        *
//Purpose : Blocks processing until USB TX buffer is available             *
//**************************************************************************
//
public sub WaitForTX()
    while ((mHIDTxIsBusy <> 0) and Attached)
        USBService()
    end while
end sub

//
//**************************************************************************
//Name    : DataAvailable                                                  *
//Purpose : Returns true if RX data is available, false otherwise          *
//**************************************************************************
//
public function DataAvailable() as boolean
    result = false
    if (mHIDRxIsBusy = 0) then
        result = true
    endif
end function

//
//**************************************************************************
//Name    : WaitForRX                                                      *
//Purpose : Blocks processing until USB RX buffer is available             *
//**************************************************************************
//
public sub WaitForRX()
    while ((mHIDRxIsBusy <> 0) and Attached)
        USBService()
    end while
end sub

//
//**************************************************************************
//Name    : ReadArray (OVERLOADED)                                         *
//Purpose : Read an array of bytes - pCount is an optional number of bytes *
//        : to read. If pCount is not given, the size of the array is used *
//**************************************************************************
// note: this routine has been overloaded and split into two to add support
// for passing an address and count
//
// read 'count' bytes into 'buf_addr' ram space. this works with any data type
public sub ReadArray(buf_addr as word, count as word)
    dim BytesRead as byte

    DisableISR()
    while ((count > 0) and Attached)
        // wait for data
        if (mHIDRxIsBusy = 0) then
            if (count > HID_REPORT_OUT_SIZE) then       // pCount > max report size
                BytesRead = HIDRxReport(buf_addr, HID_REPORT_OUT_SIZE)
                dec(count, BytesRead)
                inc(buf_addr, BytesRead)
            else                                        // pCount <= max report size
                BytesRead = HIDRxReport(buf_addr, count.byte0)
                dec(count.byte0, BytesRead)
                inc(buf_addr, BytesRead)
            endif
        endif
        _service()
    end while
    EnableISR()
end sub

public sub ReadArray(byref pBuffer() as byte, pCount as word = 0)
    if (pCount = 0) then
       pCount = bound(pBuffer) + 1
    endif
    ReadArray(addressof(pBuffer), pCount)
end sub

//
//**************************************************************************
//Name    : WriteArray (OVERLOADED)                                        *
//Purpose : Write an array of bytes - pCount is an optional number of      *
//        : to write. If pCount is not given, the array size is used       *
//        : Bytes are transfered in HIDReportBytesIn sized packets - 8, 16 *
//        : 32 or 64 bytes. If you pass an array which is only dimensioned *
//        : 4 bytes (for example, DIM Array(4) AS BYTE  then a             *
//        : 'HIDReportBytesIn' sized data packet is still transferred.     *
//        : This isn't really a problem, as the host application should    *
//        : know what to do with the data anyway.                          *
//**************************************************************************
// note: this routine has been overloaded and split into two to add support
// for passing an address and count
//
// write 'len' bytes from 'buf_addr' ram space. this works with any data type
public sub WriteArray(buf_addr as word, len as word)
    DisableISR()
    while ((len > HID_REPORT_IN_SIZE) and Attached)
        WaitForTX()
        HIDTxReport(buf_addr, HID_REPORT_IN_SIZE)
        dec(len, HID_REPORT_IN_SIZE)
        inc(buf_addr, HID_REPORT_IN_SIZE)
        _service()
    end while

    if (len.byte0 > 0) then
        WaitForTX()
        HIDTxReport(buf_addr, HID_REPORT_IN_SIZE)
        _service()
    endif
    EnableISR()
end sub

public sub WriteArray(byref pBuffer() as byte, pLength as word = 0)
    if (pLength = 0) then
        pLength = bound(pBuffer) + 1
    endif
    WriteArray(addressof(pBuffer), pLength)
end sub

//
//**************************************************************************
//Name    : ReadReport                                                     *
//Purpose : Read a HID report into RXReport buffer                         *
//**************************************************************************
//**************************************************************************
//Name    : WriteReport                                                    *
//Purpose : Write the TXReport buffer                                      *
//**************************************************************************
//
// ReadReport() and WriteReport() need to use buffers that are the same size 
// as the defined #options HID_REPORT_BYTES_IN and HID_REPORT_BYTES_OUT so
// that they match the descriptor report sizes.
//
// In order to allow the user to define their own custom report structures,
// we create a union of the user's report structures and byte arrays of
// the req'd size (in previous versions this was done by overlaying the 
// structures using ABSOLUTE declarations). After this, we then define
// aliases to the user's overlayed custom structures (TXReport and RxReport)
// so that the original syntax can be used to access the report data
//
// To define custom reports, create a .bas file containing the following public
// structure definitions...
{
    module UserReports

    // TX report
    public structure TTXReport
        ' <<add your definitions here>>
    end structure

    // RX report
    public structure TRXReport
        ' <<add your definitions here>>
    end structure

    end module
}
// Then, in your main program set '#option USER_CUSTOM_HID_REPORTS' to the filename 
// of your custom reports file before including usbhid.bas, ie
//    #option USER_CUSTOM_HID_REPORTS = "MyReportFilename.bas" 
//    include "usbhid.bas"
//

//
//-------------------------------------------------------------
// USB report structures
//-------------------------------------------------------------
// check to see if the user has defined a custom report file
#if isoption(USER_CUSTOM_HID_REPORTS)
  include USER_CUSTOM_HID_REPORTS
#else
  // no user defined report structures...
  // define some generic byte arrays to satisfy the overlay definitions 
  public structure TTXReport
      b(HID_REPORT_IN_SIZE) as byte
  end structure

  public structure TRXReport
      b(HID_REPORT_OUT_SIZE) as byte
  end structure
#endif      // USER_CUSTOM_HID_REPORTS

//
//-------------------------------------------------------------
// report buffer overlays
//-------------------------------------------------------------
//
// define types to overlay user structures with report buffers 
public structure TTXReportBuffer
    b(HID_REPORT_IN_SIZE) as byte union
    txr_user as TTXReport union         // user-defined structure
end structure

public structure TRXReportBuffer
    b(HID_REPORT_OUT_SIZE) as byte union
    rxr_user as TRXReport union         // user-defined structure
end structure

// define the overlayed structures themselves
dim 
    txrb as TTXReportBuffer,
    rxrb as TRXReportBuffer

// create public aliases to the user-defined report structures
public dim 
    TXReport as txrb.txr_user,
    RXReport as rxrb.rxr_user

//
//-------------------------------------------------------------
//Name    : ReadReport
//Purpose : Read a HID report into RXReport buffer
//-------------------------------------------------------------
//
public sub ReadReport()
    HID.WaitForRX()
    HIDRxReport(addressof(RXReport), HID_REPORT_OUT_SIZE)
  #if (USB_SERVICE = false)
    HID.USBService()
  #endif
end sub

//
//-------------------------------------------------------------
//Name    : WriteReport
//Purpose : Write the TXReport buffer
//-------------------------------------------------------------
//
public sub WriteReport()
    HID.WaitForTX()
    HIDTxReport(addressof(TXReport), HID_REPORT_IN_SIZE)
  #if (USB_SERVICE = false)
    HID.USBService()
  #endif
end sub

//
//**************************************************************************
//Name    : Initialize                                                     *
//Purpose : Initialize the USB module                                      *
//**************************************************************************
//
public sub Initialize()
    // setup usb clock (see USBClock.bas)
    USBClockInit()

    mInitializeUSBDriver()

  #if (USB_SERVICE)
    #if (USB_SERVICE_PRIORITY = ipLow)
      USB_IP = 0
    #else
      USB_IP = 1
    #endif
    USB_IF = 0      // clear USBIF
    USB_IE = 1      // enable USB interrupt

    _service()
    UIE = UIE_ENABLE_ALL    // unmask all USB interrupts
    UEIE = UEIE_ENABLE_ALL  // unmask all USB error interrupts
    enable(USB_intr_handler)
  #endif    // USB_SERVICE
end sub

//
//**************************************************************************
// initialize the module
//**************************************************************************
//
#if (USB_AUTO_INITIALIZE)
  Initialize()
#elseif (SHOW_USB_BUILD_MSGS)
  #warning "USB_AUTO_INITIALIZE false... driver NOT initialized at startup"
#endif

end module
