//
//***************************************************************************
// Name    : USBClock.bas
// Author  : Jerry Messina
// Date    : 1/21/2015
// Version : 1.0.0
//         : - initial release (code moved from usbsfr.bas)
//         : - supports #option to allow user-defined clock setup
//         : - supports #option to allow removing any delayms calls
// Notes   : USBclock
//         : device-specific clock init for USB
//***************************************************************************
//
module USBClock

// this option lets you remove any 'delayms()' calls from the clock setup
// if removed you may need to add these delays at the application level
#option USB_CLOCK_USE_DELAYMS = true

#if (USB_CLOCK_USE_DELAYMS)     // use normal 'delayms' calls
  macro _delayms(x)
    delayms(x)
  end macro
#else                           // map calls out 
  macro _delayms(x)
  end macro
#endif

//
// check to see if the user is supplying their own usb clock init routine
// if so, include it here and add an alias so that it can be found
// otherwise, just use the default routine supplied here
//
// to use your own routine, create a module containing the following:
//
//      public sub userUSBClockInit()
//          << add your clock code here >>
//      end sub
//
// and set the filename option in your main project file
//    ie #option USER_USB_CLOCK_INIT = "MyClockInit.bas"
// calls to USBClockInit() will then be redirected to userUSBClockInit()
//
#if isoption(USER_USB_CLOCK_INIT)
  // include users file
  include USER_USB_CLOCK_INIT
  // create an alias for the user's clock init routine
  public dim USBClockInit as userUSBClockInit
#else           // use default clock init routine here 
  #if isoption(SHOW_USB_BUILD_MSGS) and (SHOW_USB_BUILD_MSGS = true)
    #warning "using default USBClockInit()"
  #endif    // SHOW_USB_BUILD_MSGS
// use following default code...

//
//------------------------------------------------------------------------
// USB OSC/PLL support
//------------------------------------------------------------------------
//
// This routine is called from the USBHID or USBCDC Initialize() function.
// It can be used to setup the cpu OSCCON register(s) and PLLEN, if required.
// If the routine is not required (as is the case for many chips), it will be
// optimized away.
//
// On some of the USB microcontrollers, the PLL will not be enabled by default,
// even if a PLL enabled oscillator configuration is selected (such as HS+PLL).
// On these devices, the PLLEN bit must be set to power up the PLL
//
public sub USBClockInit()
  #if _device in (18F2450, 18F4450)
    // nothing req'd

  #elseif _device in (18F2455, 18F2550, 18F4455, 18F4550, 18F2458, 18F2553, 18F4458, 18F4553)
    // nothing req'd

  #elseif _device in (18F24J50, 18F25J50, 18F26J50, 18F44J50, 18F45J50, 18F46J50)
    // enable the PLL and wait 2ms for the PLL to lock before enabling USB module
    OSCTUNE.bits(6) = 1      // set PLLEN bit
    _delayms(2)

  #elseif _device in (18F26J53, 18F27J53, 18F46J53, 18F47J53)
    // enable the PLL and wait 2ms for the PLL to lock  before enabling USB module
    OSCTUNE.bits(6) = 1      // set PLLEN bit
    _delayms(2)

  #elseif _device in (18F65J50, 18F66J50, 18F66J55, 18F67J50, 18F85J50, 18F86J50, 18F86J55, 18F87J50)
    // enable the PLL and wait 2ms for the PLL to lock  before enabling USB module
    OSCTUNE.bits(6) = 1      // set PLLEN bit
    _delayms(2)

  #elseif _device in (18F13K50, 18F14K50)
    // nothing req'd

  #elseif _device in (18F65J94, 18F66J94, 18F67J94, 18F85J94, 18F86J94, 18F87J94, 18F95J94, 18F96J94, 18F97J94) or _
          _device in (18F66J99, 18F86J99, 18F96J99)
    // this code assumes the use of the internal FRC clock and active clock tuning
    // you must set CONFIG FOSC = FRCPLL

    // OSCCON osc control reg
    // NOSC<2:0> = 001      FRC 8MHz osc, divided by N, with PLL module
    OSCCON = %001

    // OSCCON2 (default)
    OSCCON2 = 0

    // OSCCON3 FRC reference clock postscaler divider bits
    // IRCF<2:0> = 001      div by 2 (default) gives 8MHz/2 = 4MHz into the 96MHz PLL
    OSCCON3 = %001

    // OSCCON4 contains the PLLEN and CPDIV bits, which are described as "USB system clock select"
    // however, looking at the 96MHz PLL block diagram, CPDIV only effects the System Clock Output,
    // as the USB Clock is fixed at div by 2 out of the 96MHz PLL
    // there is a fixed divide by 1.5 stage after the CPDIV
    // 7-6  CPDIV<1:0> = 00  input clock/1
    // 5    PLLEN
  #if (_clock = 64)
    OSCCON4 = %00100000             // CPDIV<1:0> = 00  system clock = Input clock/1 = 96/1/1.5=64MHz
  #elseif (_clock = 32)
    OSCCON4 = %01100000             // CPDIV<1:0> = 01  system clock = Input clock/2 = 96/2/1.5=32MHz
  #else
    #error "unsupported 'clock' setting"
  #endif

    // Enable Active Clock Tuning
    ACTCON.bits(4) = 1              // set ACTSRC to use USB host clock
    ACTCON.bits(7) = 1              // set ACTEN to enable ACT module

    // Wait for the PLL to lock
    _delayms(2)
    while OSCCON2.bits(5) = 0       // test PLL LOCK bit
    end while

  #elseif _device in (18F24K50, 18F25K50, 18F45K50)
    // this code assumes the use of the internal FRC clock and active clock tuning
    // you must set CONFIG registers to match

    // turn off PLL for safety while we setup clock
    OSCCON2.bits(4) = 0             // PLLEN

    OSCCON = (%111 << 4) + %000     // PRIMARY_OSC 48Mhz (16MHz + 3xPLL)
    OSCTUNE.bits(7) = 1             // SPLLMULT 3xPLL
    OSCCON2.bits(4) = 1             // PLLEN

    // Enable Active Clock Tuning
    ACTCON.bits(4) = 1              // set ACTSRC to use USB host clock
    ACTCON.bits(7) = 1              // set ACTEN to enable ACT module

    // wait a bit for the clock to stabilize
    _delayms(2)

  #else
    // add new device USB clock setup here
  #endif    // _device in
end sub
#endif      // isoption(USER_USB_CLOCK_INIT)

end module
