//
//***************************************************************************
// Name    : usbmem.bas
// Author  : Jerry Messina
// Notice  : Copyright (c) 2011
//         : All Rights Reserved
// Date    : 1/2/2015
// Version : 1.4.5
//         : - use SF protected blocks (remove extended ram definitions)
//         : - add K50 and J94 families (requires MPASM v5.49 or later)
//         : 1.4.3
//         : - change extended ram size and extended ram base for 128 bytes
//         :   of ext ram on 14K50 and 2450, giving 384 bytes of user ram
// Version : 1.4 initial release
// Notes   : USB Memory Map
//         : device-specific memory sizes and locations
//
// This file supports all 18F USB devices from MPLAB v8.92/MPASM v5.49
// (45 devices total) including:
// 18F2450, 18F2455, 18F2458, 18F2550, 18F2553, 18F4450, 
// 18F4455, 18F4458, 18F4550, 18F4553, 
// 18F13K50, 18F14K50, 18F24K50, 18F25K50, 18F45K50, 
// 18F24J50, 18F25J50, 18F26J50, 18F44J50, 18F45J50, 18F46J50, 
// 18F65J50, 18F66J50, 18F67J50, 18F85J50, 18F86J50, 18F87J50, 
// 18F66J55, 18F86J55, 
// 18F26J53, 18F27J53, 18F46J53, 18F47J53, 
// 18F65J94, 18F66J94, 18F67J94, 18F85J94, 18F86J94, 18F87J94, 
// 18F95J94, 18F96J94, 18F97J94, 18F66J99, 18F86J99, 18F96J99
//
// MPLAB v8.92 is available from the Microchip DevTools archive page
// and can currently be found at:
// http://ww1.microchip.com/downloads/en/DeviceDoc/MPLAB_IDE_8_92.zip
//
//***************************************************************************
//
module usbmem

//
//---------------------------------------------------------------------------
// USB device summary
//---------------------------------------------------------------------------
//
//  device      flash       ram             usb(BDT)        EPx endpoints****
//  18F2450     16K         512(000-1FF)    256(400-4FF)*** 64(0-63)    16
//  18F4450     16K         512(000-1FF)    256(400-4FF)    64(0-63)    16
//
//  18Fxx50*    24-32K      2K(000-3FF)     1K(400-4FF)*    64(0-63)    16
//  18F4x50     24-32K      2K(000-3FF)     1K(400-4FF)*    64(0-63)    16
//
//  18FxxJ50    16-64K      3776(000-EBF)   all(400-4FF)**  64(0-63)    16
//
//  18F87J50    32-128K     3904(000-F3F)   all(400-4FF)**  64(0-63)    16
//
//  18FxxJ53    64-128K     3776(000-EBF)   all(D00-DFF)**  64(0-63)    16
//
//  18F13K50    8K          256(000-0FF)    256(200-2FF)*** 32(0-31)    8
//  18F14K50    16K         512(000-1FF)    256(200-2FF)    32(0-31)    8
//
//  18FxxK50    16-32K      2K(000-3FFF)    1K(400-4FF)*    64(0-63)    16
//
//  18FxxJ94/99 32-128K     3578(000-DFA)   all(100-1FF)    64(0-63)    16
//
//  The ram sizes include the entire ram available (_maxram) excluding the SFR's.
//  For USB applications you must account for USB BDT and buffer requirements.
//
//  * includes 2455, 2550, 4455, 4550, 2458, 2553, 4458, 4553
//    these parts differ only in flash size and 10/12-bit ADC
//  * BDT is at 400-4FF. USB data must be in the USB ram space (400-7FF).
//
//  ** for the 'J' series parts, only the BDT must reside in the USB page.
//     the USB data can be located anywhere in ram space, and can be any size up
//     to the SFR at the top of memory
//
//  *** for the 2450 and 13K/14K50 parts, the BDT and USB data must be contained
//      in the 256 byte USB bank (400/200). For the 1xK50, the BDT must fit into
//      200h-27fh.
//
//  *** the EPx endpoints are bidirectional in/out pairs
//

//
//---------------------------------------------------------------------------
// memory usage
//---------------------------------------------------------------------------
//
// The BDT must reside here, along with the descriptors and usually the usb 
// endpoint buffers. this is fixed address that varies depending on the device,
// and cannot be changed
//   #define USB_RAM_BASE = $400     // base addr of the USB BDT ($200, $400, $D00, $100)

// this is the max amount of ram reserved for the SF protected_a region
//   #option USB_PROTECTED_BLOCKSIZE = 256

// this is the amount of ram actually reserved from SF allocation
// it defaults to USB_PROTECTED_BLOCKSIZE, but can be resized depending on the
// '#option USBMEM_RESERVE_MEMORY' setting. the default setting will allow
// the USBCDC/USBHID modules to resize the region so that only the reqd amount
// of ram is reserved, freeing up the rest for the application
//   #variable USB_RAM_SIZE = USB_PROTECTED_BLOCKSIZE     // USB ram, in bytes

// the SF IDE will account for the ram memory that is reserved using the 
// '_protected_xxxx' feature... it will not be displayed in the summary
// 'xxx variable bytes used' status line, but it will show up in the maxram
// portion of the Results window message '...bytes used from a possible <maxram>'.
// the IDE will reduce the <maxram> by the amount reserved by the protected region.
// for example, if we have a device with 2048 bytes of total ram  and reserve 
// 256 bytes for the USB BDT and buffers, that leaves a max of 2048-256 = 1792 
// bytes available, and we get the message 'xxxx bytes used from a possible 1792'.
// you can also check the .idf file '.ram_used' value in the .header section

//
//---------------------------------------------------------------------------
// USB memory definitions
//---------------------------------------------------------------------------
//
// this is the default amount of ram reserved for the SF protected_a region
#option USB_PROTECTED_BLOCKSIZE = 256

// number of bytes of USB memory to reserve (this can be changed later on)
#variable USB_RAM_SIZE = USB_PROTECTED_BLOCKSIZE       // USB ram, in bytes

// set the USB RAM base
#if _device in (18F2450, 18F4450)
    // 768 bytes of total ram (includes 256 bytes of dual-port USB ram)
    // however, there is only 512 bytes available to SF ($000-$1FF)
    // since the USB ram is not contiguous
    #define USB_RAM_BASE = $400                     // fixed BDT + descriptors + buffers

#elseif _device in (18F2455, 18F2550, 18F4455, 18F4550, 18F2458, 18F2553, 18F4458, 18F4553)
    // 2K of ram (includes 1K of dual-port USB ram)
    #define USB_RAM_BASE = $400                     // fixed BDT + descriptors + buffers

#elseif _device in (18F24J50, 18F25J50, 18F26J50, 18F44J50, 18F45J50, 18F46J50)
    // 3776 bytes of total ram (USB ram can be anywhere)
    #define USB_RAM_BASE = $400                     // fixed BDT + descriptors + buffers

#elseif _device in (18F26J53, 18F27J53, 18F46J53, 18F47J53)
    // 3.8K bytes of total ram (USB ram can be anywhere)
    #define USB_RAM_BASE = $D00                     // fixed BDT + descriptors + buffers

#elseif _device in (18F65J94, 18F66J94, 18F67J94, 18F85J94, 18F86J94, 18F87J94, 18F95J94, 18F96J94, 18F97J94) or _
        _device in (18F66J99, 18F86J99, 18F96J99)
    // 3.8K bytes of total ram (USB ram can be anywhere)
    #define USB_RAM_BASE = $100                     // fixed BDT (**DATASHEET INCORRECT**)

#elseif _device in (18F65J50, 18F66J50, 18F66J55, 18F67J50, 18F85J50, 18F86J50, 18F86J55, 18F87J50)
    // 3904 bytes of total ram (USB ram can be anywhere)
    #define USB_RAM_BASE = $400                     // fixed BDT + descriptors + buffers

#elseif _device in (18F13K50, 18F14K50) then
    // 13K50 - 256 bytes of total ram (does not include 256 bytes of dual-port USB ram)
    // 14K50 - 768 bytes of total ram (includes 256 bytes of dual-port USB ram)
    // make sure _maxram from the device .bas file is correct for the part (some versions are not)
    #if _device in (18F13K50)
      #variable _maxram = $100              // 256 bytes of ram (not including USB ram)
    #elseif _device in (18F14K50)
      #variable _maxram = $300              // 768 bytes of ram
    #endif
    #define USB_RAM_BASE = $200             // fixed BDT + descriptors + buffers

#elseif _device in (18F24K50, 18F25K50, 18F45K50)
    // 2K of ram (includes 1K of dual-port USB ram)
    #define USB_RAM_BASE = $400             // fixed BDT + descriptors + buffers

#else
  // add new devices here
  #error "unsupported device"
#endif

//
// the memory region starting at USB_RAM_BASE must be reserved from normal SF allocation.
// if option USBMEM_RESERVE_MEMORY = false (the default setting), then we don't reserve the 
// block here... we let USBCDC/USBHID resize the block and only reserve what's actually required.
// if option USBMEM_RESERVE_MEMORY = true, then we reserve the entire USB_RAM_SIZE block here
// and it remains a fixed size
//
#option USBMEM_RESERVE_MEMORY = false
#if (USBMEM_RESERVE_MEMORY)
  #if isoption(SHOW_USB_BUILD_MSGS) and (SHOW_USB_BUILD_MSGS = true)
    #warning "USBMEM allocating fixed USB_RAM_SIZE block"
  #endif
  // create protected block to reserve USB ram from SF allocation
  #define _protected_a_start = USB_RAM_BASE
  #define _protected_a_end   = USB_RAM_BASE + (USB_RAM_SIZE - 1)
#endif      // USBMEM_RESERVE_MEMORY

end module
